//=============================================================================
//
//   This file is part of the SSEX text editor distribution
//   Copyright (C) 2001 Szymon Stefanek (stefanek@tin.it)
//
//   This program is FREE software. You can redistribute it and/or
//   modify it under the terms of the GNU General Public License
//   as published by the Free Software Foundation; either version 2
//   of the License, or (at your opinion) any later version.
//
//   This program is distributed in the HOPE that it will be USEFUL,
//   but WITHOUT ANY WARRANTY; without even the implied warranty of
//   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
//   See the GNU General Public License for more details.
//
//   You should have received a copy of the GNU General Public License
//   along with this program. If not, write to the Free Software Foundation,
//   Inc. ,59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
//
//=============================================================================

#include "editor.h"
#include "kvi_options.h"

#undef Unsorted


#include <qframe.h>
#include <qpainter.h>
#include <qfontmetrics.h>
#include <qfile.h>
#include <qnamespace.h>
#include <qcursor.h>
#include <qclipboard.h>
#include <qapplication.h>
#include <qfiledialog.h>
#include <qfileinfo.h>
#include <qmessagebox.h>
#include <qlayout.h>
#include <qtoolbutton.h>
#include <qlabel.h>
#include <qinputdialog.h>
#include <qfontdialog.h>
#include <qregexp.h>
#include <qcolordialog.h>
#include <qfontdialog.h>
#include <qvbox.h>
#include <qdatastream.h>
#include <qdir.h>

#include <ctype.h>
#include <string.h>

#ifdef KVIRC_SSEX
	#include "kvi_settings.h"
#endif

#ifdef COMPILE_ON_WINDOWS
	#define strncasecmp _strnicmp
	#define strcasecmp _stricmp
#endif

#define __DEBUG


#ifdef __DEBUG
	#define __assert(_expr) if(! (_expr) )if(!(_expr))debug("RANGE ASSERT : \"%s\" is false in %s (%d)",#_expr,__FILE__,__LINE__);
#else
	#define __assert(_expr)
#endif

static SSEXOptions * g_pSSEXOptions = 0;


SSEXCompletion::SSEXCompletion()
{
	m_pTokens = new KviPtrList<QCString>;
}

SSEXCompletion::~SSEXCompletion()
{
	delete m_pTokens;
}

QCString SSEXCompletion::getCompletion(const char * tok)
{
	if(m_pTokens->isEmpty())return QCString();

	QCString match;

	
	int wLen = strlen(tok);
	bool bStart = true;

	for(QCString * s = m_pTokens->first();s;s = m_pTokens->next())
	{
		if(strncmp(s->data(),tok,wLen) == 0)
		{
			if(bStart)
			{
				match = *s;
				bStart = false;
			}
			if(s->length() < match.length())match.resize(s->length() + 1);
			// All the matches here have length >= word.len()!!!
			char * c1 = s->data() + wLen;
			char * c2 = match.data() + wLen;
			while(*c1 && (*c1 == *c2))c1++,c2++;
			int len = c1 - s->data();
			if(len < match.length())match.resize(len + 1);
		}
	}

	match.remove(0,wLen);
	return match;
}

void SSEXCompletion::tokenize(const QCString &data)
{
	const char * aux = data.data();
	const char * begin = 0;
	while(*aux)
	{
		if(isalnum(*aux) || (*aux == '_'))
		{
			if(!begin)begin=aux;
		} else {
			if(begin)
			{
				if(begin != aux)
				{
					m_pTokens->append(new QCString(begin,(aux - begin) + 1));
				}
			}
			begin = 0;
		}
		aux++;
	}
	if(begin)
	{
		m_pTokens->append(new QCString(begin,(aux - begin) + 1));
	}
}


SSEXColorEditor::SSEXColorEditor(SSEXOptionsDialog * dlg,QWidget * par,QColor * pClr,const char * txt)
: QHBox(par)
{
	m_pClr = pClr;
	m_actualClr = *pClr;
	QLabel * l = new QLabel(txt,this);
	m_pButton = new QPushButton(this);
	m_pButton->setPalette(QPalette(*pClr));
	m_pButton->setMinimumWidth(120);
	m_pButton->setMaximumWidth(120);
	connect(m_pButton,SIGNAL(clicked()),this,SLOT(chooseColor()));
	connect(dlg,SIGNAL(doCommit()),this,SLOT(commit()));
}


SSEXColorEditor::~SSEXColorEditor()
{
}

void SSEXColorEditor::chooseColor()
{
	QColor clr = QColorDialog::getColor(*m_pClr);
	if(clr.isValid())
	{
		m_actualClr = clr;
		m_pButton->setPalette(QPalette(clr));
	}
}

void SSEXColorEditor::commit()
{
	*m_pClr = m_actualClr;
}

SSEXFontEditor::SSEXFontEditor(SSEXOptionsDialog * dlg,QWidget * par,QFont * pFont,const char * txt)
: QHBox(par)
{
	m_pFont = pFont;
	QLabel * l = new QLabel(txt,this);
	m_pButton = new QPushButton(this);
	m_pButton->setFont(*pFont);
	m_pButton->setText(pFont->family());
	m_pButton->setMinimumWidth(120);
	m_pButton->setMaximumWidth(120);
	connect(m_pButton,SIGNAL(clicked()),this,SLOT(chooseFont()));
	connect(dlg,SIGNAL(doCommit()),this,SLOT(commit()));
}


SSEXFontEditor::~SSEXFontEditor()
{
}

void SSEXFontEditor::chooseFont()
{
	bool bOk = false;
	QFont fnt = QFontDialog::getFont(&bOk,m_pButton->font());
	if(bOk)
	{
		m_pButton->setFont(fnt);
		m_pButton->setText(fnt.family());
	}
}

void SSEXFontEditor::commit()
{
	*m_pFont = m_pButton->font();
}


SSEXOptionsDialog::SSEXOptionsDialog()
: QTabDialog()
{
	setOkButton("Ok");
	setCancelButton("Cancel");
	connect(this,SIGNAL(cancelButtonPressed()),this,SLOT(cancelClicked()));
	connect(this,SIGNAL(applyButtonPressed()),this,SLOT(okClicked()));

	SSEXColorEditor * ed;
	SSEXFontEditor * fe;
	QLabel * l;

	QVBox * page = new QVBox(this);
	page->setMargin(5);

	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlBackground),"Background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlDisabledBackground),"Disabled background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlExtBackground),"External background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlMarkBackground),"Mark background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlNormalText),"NormalText");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlCursor),"Cursor");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlComment),"Comment");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlTab),"Tab");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlVariable),"Variable");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlOperator),"Operator");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrPerlBrace),"Brace");
	fe = new SSEXFontEditor(this,page,&(g_pSSEXOptions->fntPerl),"Font");

	addTab(page,"KVS");

	page = new QVBox(this);
	page->setMargin(5);

	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrTxtBackground),"Background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrTxtDisabledBackground),"Disabled background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrTxtExtBackground),"External background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrTxtMarkBackground),"Mark background");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrTxtNormalText),"Foreground");
	ed = new SSEXColorEditor(this,page,&(g_pSSEXOptions->clrTxtCursor),"Cursor");

	fe = new SSEXFontEditor(this,page,&(g_pSSEXOptions->fntTxt),"Font");

	l = new QLabel("",page);

	page->setStretchFactor(l,1);

	addTab(page,"Plain text");

	page = new QVBox(this);
	page->setMargin(5);
}

SSEXOptionsDialog::~SSEXOptionsDialog()
{
}

void SSEXOptionsDialog::showEvent(QShowEvent *e)
{
	move((QApplication::desktop()->width() - width()) / 2,(QApplication::desktop()->height() - height()) / 2);
	QDialog::showEvent(e);
}

void SSEXOptionsDialog::closeEvent(QCloseEvent *e)
{
	e->ignore();
	g_pSSEXOptions->optionsDialogTerminated();
}

void SSEXOptionsDialog::okClicked()
{
	emit doCommit();
	g_pSSEXOptions->optionsDialogTerminated();
	g_pSSEXOptions->apply();
}

void SSEXOptionsDialog::cancelClicked()
{
	g_pSSEXOptions->optionsDialogTerminated();
}



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////
//// SSEXFindWidget
////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


SSEXFindWidget::SSEXFindWidget(SSEXEditor * parent)
:QFrame(parent)
{
	setFrameStyle(QFrame::WinPanel | QFrame::Raised);
	m_pEditor = parent;
	setCursor(arrowCursor);

	QGridLayout *g = new QGridLayout(this,13,3,4,0);
	QToolButton *tb = new QToolButton(DownArrow,this);
	tb->setFixedSize(12,12);
	tb->setAutoRepeat(false);
	connect(tb,SIGNAL(clicked()),m_pEditor,SLOT(toggleFindWidget()));
	g->addWidget(tb,0,2);

	QFrame *f = new QFrame(this);
	f->setFrameStyle(QFrame::HLine | QFrame::Sunken);
	g->addMultiCellWidget(f,1,1,0,2);

	QLabel *l = new QLabel("String to find :",this);
	g->addMultiCellWidget(l,2,2,0,2);

	m_pFindStringEdit = new QLineEdit(this);
	g->addMultiCellWidget(m_pFindStringEdit,3,3,0,2);
	
	setFocusProxy(m_pFindStringEdit);

	m_pRegExpCheckBox = new QCheckBox("Regular expression",this);
	g->addMultiCellWidget(m_pRegExpCheckBox,4,4,0,2);

	m_pCaseSensitiveCheckBox = new QCheckBox("Case sensitive",this);
	m_pCaseSensitiveCheckBox->setChecked(true);
	g->addMultiCellWidget(m_pCaseSensitiveCheckBox,5,5,0,2);


	QPushButton *b = new QPushButton("Find next",this);
	b->setDefault(true);
	g->addWidget(b,6,0);
	connect(b,SIGNAL(clicked()),this,SLOT(findNextClicked()));

	b = new QPushButton("Find prev",this);
	g->addMultiCellWidget(b,6,6,1,2);
	connect(b,SIGNAL(clicked()),this,SLOT(findPrevClicked()));

	f = new QFrame(this);
	f->setFrameStyle(QFrame::HLine | QFrame::Sunken);
	g->addMultiCellWidget(f,7,7,0,2);

	l = new QLabel("Replacement text :",this);
	g->addMultiCellWidget(l,8,8,0,2);

	m_pReplaceStringEdit = new QLineEdit(this);
	g->addMultiCellWidget(m_pReplaceStringEdit,9,9,0,2);

	m_pReplace = new QPushButton("Replace",this);
	connect(m_pReplace,SIGNAL(clicked()),m_pEditor,SLOT(replace()));
	g->addWidget(m_pReplace,10,0);
	m_pReplace->setEnabled(m_pEditor->hasSelection());

	b = new QPushButton("Replace all",this);
	connect(b,SIGNAL(clicked()),this,SLOT(replaceAllClicked()));
	g->addMultiCellWidget(b,10,10,1,2);

	m_pReplaceAndFindNext = new QPushButton("Replace and find next",this);
	connect(m_pReplaceAndFindNext,SIGNAL(clicked()),this,SLOT(replaceAndFindNextClicked()));
	g->addMultiCellWidget(m_pReplaceAndFindNext,11,11,0,2);
	m_pReplaceAndFindNext->setEnabled(m_pEditor->hasSelection());

	m_pReplaceAllInSelection = new QPushButton("Replace all in selection",this);
	connect(m_pReplaceAllInSelection,SIGNAL(clicked()),this,SLOT(replaceAllInSelectionClicked()));
	g->addMultiCellWidget(m_pReplaceAllInSelection,12,12,0,2);
	m_pReplaceAllInSelection->setEnabled(m_pEditor->hasSelection());


	g->setResizeMode(QGridLayout::Fixed);
}

SSEXFindWidget::~SSEXFindWidget()
{
}

void SSEXFindWidget::findNextClicked()
{
	if(m_pRegExpCheckBox->isChecked())m_pEditor->findNextRegExp();
	else m_pEditor->findNext();
}

void SSEXFindWidget::findPrevClicked()
{
	if(m_pRegExpCheckBox->isChecked())m_pEditor->findPrevRegExp();
	else m_pEditor->findPrev();
}

void SSEXFindWidget::replaceAllClicked()
{
	if(m_pRegExpCheckBox->isChecked())m_pEditor->replaceAllRegExp();
	else m_pEditor->replaceAll();
}

void SSEXFindWidget::replaceAndFindNextClicked()
{
	if(m_pRegExpCheckBox->isChecked())m_pEditor->replaceAndFindNextRegExp();
	else m_pEditor->replaceAndFindNext();
}

void SSEXFindWidget::replaceAllInSelectionClicked()
{
	if(m_pRegExpCheckBox->isChecked())m_pEditor->replaceAllInSelectionRegExp();
	else m_pEditor->replaceAllInSelection();
}

void SSEXFindWidget::mousePressEvent(QMouseEvent *e)
{
	m_pressPoint = e->pos();
}

void SSEXFindWidget::mouseMoveEvent(QMouseEvent *)
{
	QPoint p=m_pEditor->mapFromGlobal(QCursor::pos());
	p-=m_pressPoint;
	if(p.x() < 0)p.setX(0);
	else if((p.x() + width()) > m_pEditor->width())p.setX(m_pEditor->width() - width());
	if(p.y() < 0)p.setY(0);
	else if((p.y() + height()) > m_pEditor->height())p.setY(m_pEditor->height() - height());
	move(p);
}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////
//// SSEXEditorOptions
////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////





SSEXOptions::SSEXOptions()
{
	m_pRefList = new KviPtrList<SSEXEditor>;
	m_pRefList->setAutoDelete(false);
	m_pOptionsDialog = 0;
	load();
}

SSEXOptions::~SSEXOptions()
{
	save();
	if(m_pOptionsDialog)delete m_pOptionsDialog;
	delete m_pRefList;
}

void SSEXOptions::edit()
{
	if(!m_pOptionsDialog)
	{
		m_pOptionsDialog = new SSEXOptionsDialog();
	}
	m_pOptionsDialog->show();
	m_pOptionsDialog->raise();
	m_pOptionsDialog->setFocus();
}

void SSEXOptions::optionsDialogTerminated()
{
	delete m_pOptionsDialog;
	m_pOptionsDialog = 0;
}

void SSEXOptions::apply()
{
	for(SSEXEditor * e = m_pRefList->first();e;e = m_pRefList->next())
	{
		e->updateOptions();
	}
}

void SSEXOptions::attach(SSEXEditor * ed)
{
	if(!g_pSSEXOptions)g_pSSEXOptions = new SSEXOptions();
	g_pSSEXOptions->attachRef(ed);
}

void SSEXOptions::detach(SSEXEditor * ed)
{
	g_pSSEXOptions->detachRef(ed);
	if(!(g_pSSEXOptions->hasRef()))
	{
		delete g_pSSEXOptions;
		g_pSSEXOptions = 0;
	}
}

void SSEXOptions::initDefaults()
{
     QFont f = QFont(KVI_OPTION_FONT(KviOption_fontIrcView));
	f.setStyleHint(QFont::TypeWriter);
	//f.setPointSize(11);
	clrTxtBackground          = QColor(0,0,0);
	clrTxtDisabledBackground  = QColor(40,40,40);
	clrTxtExtBackground       = QColor(40,40,40);
	clrTxtMarkBackground      = QColor(25,25,45);
	clrTxtNormalText          = QColor(50,255,0);
	clrTxtCursor              = QColor(255,0,0);
	fntTxt                    = f;

	clrPerlBackground         = QColor(0,0,0);
	clrPerlDisabledBackground = QColor(40,40,40);
	clrPerlExtBackground      = QColor(40,40,40);
	clrPerlMarkBackground     = QColor(25,25,45);
	clrPerlNormalText         = QColor(20,255,20);
	clrPerlCursor             = QColor(255,0,0);
	clrPerlComment            = QColor(35,180,0);
	clrPerlTab                = QColor(70,70,70);
	clrPerlVariable           = QColor(255,255,50);
	clrPerlOperator           = QColor(150,150,40);
	clrPerlBrace              = QColor(255,0,0);
	fntPerl                   = f;

	uUndoDepth                = 256;
	uTabsNumSpaces            = 4;
}

#define SSEXOPTIONSVERSION "ssexoptions2.0.0"

void SSEXOptions::load()
{
	QString thedir = QDir::homeDirPath();
	thedir += "/.kvircssexeditorrc";
	QFile f(thedir);
	
	if(!f.open(IO_ReadOnly))
	{
		initDefaults();
		return;
	}

	QDataStream s(&f);

	QString tmp;
	s >> tmp;
	if(tmp != SSEXOPTIONSVERSION)
	{
		initDefaults();
		return;
	}

	s >> clrTxtBackground;
	s >> clrTxtDisabledBackground;
	s >> clrTxtExtBackground;
	s >> clrTxtMarkBackground;
	s >> clrTxtNormalText;
	s >> clrTxtCursor;
	s >> fntTxt;

	s >> clrPerlBackground;
	s >> clrPerlDisabledBackground;
	s >> clrPerlExtBackground;
	s >> clrPerlMarkBackground;
	s >> clrPerlNormalText;
	s >> clrPerlCursor;
	s >> clrPerlComment;
	s >> clrPerlTab;
	s >> clrPerlVariable;
	s >> clrPerlOperator;
	s >> clrPerlBrace;
	s >> fntPerl;

	s >> uUndoDepth;
	s >> uTabsNumSpaces;


	f.close();
}

void SSEXOptions::save()
{
	QString thedir = QDir::homeDirPath();
	thedir += "/.kvircssexeditorrc";
	QFile f(thedir);

	if(!f.open(IO_WriteOnly))return;

	QDataStream s(&f);

	QString tmp = SSEXOPTIONSVERSION;

	s << tmp;
	s << clrTxtBackground;
	s << clrTxtDisabledBackground;
	s << clrTxtExtBackground;
	s << clrTxtMarkBackground;
	s << clrTxtNormalText;
	s << clrTxtCursor;
	s << fntTxt;

	s << clrPerlBackground;
	s << clrPerlDisabledBackground;
	s << clrPerlExtBackground;
	s << clrPerlMarkBackground;
	s << clrPerlNormalText;
	s << clrPerlCursor;
	s << clrPerlComment;
	s << clrPerlTab;
	s << clrPerlVariable;
	s << clrPerlOperator;
	s << clrPerlBrace;
	s << fntPerl;

	s << uUndoDepth;
	s << uTabsNumSpaces;

	f.close();

}

////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////
//// SSEXEditor
////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////



#define SSEX_EDITOR_BORDER 5
#define SSEX_EDITOR_BLINK_TIME 500
#define SSEX_EDITOR_DRAG_TIMEOUT 150

SSEXEditor::SSEXEditor(QWidget * parent)
:SSEXTableView(parent)
{
	SSEXOptions::attach(this);

	setFrameStyle(QFrame::Panel|QFrame::Sunken);
	setNumRows(1);

	setCursor(ibeamCursor);

	m_pLines = new KviPtrList<SSEXEditorTextLine>;
	m_pLines->setAutoDelete(true);

	m_mode = Normal;

	m_iCursorRow              = 0;
	m_iCursorPosition         = 0;
	m_iCursorPositionInPixels = 0;
	m_iLastCursorRow          = 0;
	m_iLastCursorPosition     = 0;

	m_iMaxTextWidth = 0;
	m_pMemBuffer = new QPixmap(cellWidth() + 1,cellHeight() + 1);
	m_pBlinkTimer = new QTimer();
	m_bCursorOn = false;
	m_bOverwrite = false;

	m_bRecordingKeystrokes = false;
	m_pKeystrokes = new KviPtrList<SSEXEditorKeystroke>;
	m_pKeystrokes->setAutoDelete(true);

	m_bHasSelection = false;

	m_bHasBlockMark = false;

	m_pContextPopup = new QPopupMenu(this);
	m_pFindPopup = new QPopupMenu(this);
	connect(m_pFindPopup,SIGNAL(aboutToShow()),this,SLOT(findPopupAboutToShow()));
	m_pEditPopup = new QPopupMenu(this);
	connect(m_pEditPopup,SIGNAL(aboutToShow()),this,SLOT(editPopupAboutToShow()));

	m_pUndoList = new KviPtrList<SSEXUndo>;
	m_pUndoList->setAutoDelete(true);

	connect(m_pBlinkTimer,SIGNAL(timeout()),this,SLOT(blinkTimer()));

	m_bModified = false;
	m_pDragTimer = new QTimer();
	connect(m_pDragTimer,SIGNAL(timeout()),this,SLOT(dragTimer()));

	updateFontDependantVariables();
	updateCellSize();

	setBackgroundMode(NoBackground);

	setFocusPolicy(QWidget::StrongFocus);

	m_iBlinkTime = SSEX_EDITOR_BLINK_TIME;
	m_szFileName = "";

	m_pFindWidget = new SSEXFindWidget(this);
	m_pFindWidget->hide();

	m_lastFindWidgetPosition = QPoint(20,20);

	initializeCurrentMode();
	setText("");

}

SSEXEditor::~SSEXEditor()
{
	closeFile(); // cancel does nothing here

	delete m_pFindWidget;
	delete m_pLines;
	delete m_pMemBuffer;
	delete m_pBlinkTimer;
	delete m_pDragTimer;
	delete m_pKeystrokes;
	delete m_pUndoList;

	SSEXOptions::detach(this);
}

static const char * m_modeTable[]=
{
	"Plain text",
	"KVS"
};

const char * SSEXEditor::modeName()
{
	switch(m_mode)
	{
		case Normal: return m_modeTable[0]; break;
		case Perl: return m_modeTable[1]; break;
		default: return m_modeTable[0]; break;
	}
}


bool SSEXEditor::closeFile()
{
	if(m_bModified)
	{
		QString msg;
		msg.sprintf("The file %s has been modified.\nDo you wish to save your changes before closing?",m_szFileName.isEmpty() ? "unnamed" : m_szFileName.data());
		int ret = QMessageBox::warning(this,"Warning - KVIrc",msg,"&Save","&Discard","Cancel");
		if(ret == 0){
			if(!saveFile())return closeFile(); //cancelled after this...repeat plz...
			else return true; //saved
		} else if(ret == 1){
			m_bModified = false; // don't call closeFile in destructor
			return true;
		} else return false; //cancelled;
	}
	return true; // safe to close
}


void SSEXEditor::getText(QCString &buffer)
{
	buffer = "";
	for(SSEXEditorTextLine *l=m_pLines->first();l;l=m_pLines->next())
	{
		buffer.append(l->text);
		buffer.append("\n");
	}
}

void SSEXEditor::markBlock()
{
	if(m_bHasBlockMark)clearBlockMark(false);

	int iPos = m_iCursorPosition;
	int iRow = m_iCursorRow;


	char chFound = 0;
	char chMatch = 0;

	int leftNormalPar = 0;
	int leftSquarePar = 0;
	int leftBrackets  = 0;
	int leftSquares   = 0;
	

	SSEXEditorTextLine * l = m_pLines->at(iRow);

	if(iPos > l->length)iPos = l->length;

	while((chFound == 0) && iRow >= 0)
	{
		iPos--;
		if(iPos < 0)
		{
			iRow--;
			if(iRow >= 0)
			{
				l = m_pLines->at(iRow);
				iPos = l->length - 1;
			}
		}
		if(iPos > -1)
		{
			switch(l->text.at(iPos))
			{
				case ')': leftNormalPar++; break;
				case '(': if(leftNormalPar == 0){ chFound = '('; chMatch = ')'; } else leftNormalPar--; break;
				case ']': leftSquarePar++; break;
				case '[': if(leftSquarePar == 0){ chFound = '['; chMatch = ']'; }else leftSquarePar--; break;
				case '}': leftBrackets++; break;
				case '{': if(leftBrackets == 0){ chFound = '{'; chMatch = '}'; } else leftBrackets--; break;
			}
		}
	}

	if(chFound)
	{
//		debug("Found left end");

		// Ok..we have the left end
		m_blockMark1 = QPoint(iPos,iRow);
		// Now do the same in the right direction

		iPos = m_iCursorPosition;
		iRow = m_iCursorRow;

		
		l = m_pLines->at(iRow);
	

		while((chFound != 0) && (iRow < numRows()))
		{
			if(iPos < l->length)
			{
				char c = l->text.at(iPos);
				if(c == chFound)leftNormalPar++;
				else if(c == chMatch)
				{
					if(leftNormalPar == 0)chFound = 0;
					else leftNormalPar--;
				}
			}
			iPos++;
			if((iPos >= l->length) && (chFound != 0))
			{
				iPos = 0;
				iRow++;
				if(iRow < numRows())l = m_pLines->at(iRow);
			}
		}

		if(chFound == 0)
		{
//			debug("Found right end");
			// Ok..have the whole mark
			m_blockMark2 = QPoint(iPos,iRow);
			m_bHasBlockMark = true;
		}

	}

	update();
}

void SSEXEditor::clearBlockMark(bool bUpdate)
{
	if(!m_bHasBlockMark)return;
	m_bHasBlockMark = false;
	if(bUpdate)update();
}

void SSEXEditor::toggleFindWidget()
{
	if(m_pFindWidget->isVisible())
	{
		m_lastFindWidgetPosition = m_pFindWidget->pos();
		m_pFindWidget->hide();
		setFocus();
	} else {
		if((m_lastFindWidgetPosition.x() + m_pFindWidget->width()) > width())m_lastFindWidgetPosition.setX(width() - m_pFindWidget->width());
		if((m_lastFindWidgetPosition.y() + m_pFindWidget->height()) > height())m_lastFindWidgetPosition.setY(height() - m_pFindWidget->height());
		m_pFindWidget->move(m_lastFindWidgetPosition);
		m_pFindWidget->show();
		m_pFindWidget->m_pFindStringEdit->setFocus();
	}
}

void SSEXEditor::setModified(bool bModified)
{
	if(bModified != m_bModified)
	{
		m_bModified = bModified;
		emit modifyStateChanged(this,m_bModified);
	}
}

void SSEXEditor::setHasSelection(bool bHasSelection)
{
	m_bHasSelection = bHasSelection;
	if(m_pFindWidget)
	{
		m_pFindWidget->m_pReplace->setEnabled(bHasSelection);
		m_pFindWidget->m_pReplaceAndFindNext->setEnabled(bHasSelection);
		m_pFindWidget->m_pReplaceAllInSelection->setEnabled(bHasSelection);
	}
}

void SSEXEditor::blinkTimer()
{
	m_bCursorOn = !m_bCursorOn;
	updateCell(m_iCursorRow);
}

void SSEXEditor::setBlinkTime(int blinkTime)
{
	if(blinkTime < 100)blinkTime = 100;
	if(blinkTime > 10000)blinkTime = 10000;
	m_iBlinkTime = 100;
	if(hasFocus() && isEnabled())startBlinking();
}

void SSEXEditor::startBlinking()
{
	if(m_pBlinkTimer->isActive())m_pBlinkTimer->stop();
	m_bCursorOn = true;
	m_pBlinkTimer->start(m_iBlinkTime);	
}


void SSEXEditor::stopBlinking()
{
	if(m_pBlinkTimer->isActive())m_pBlinkTimer->stop();
	if(m_bCursorOn){
		m_bCursorOn = false;
		updateCell(m_iCursorRow);
	}
}

void SSEXEditor::focusInEvent(QFocusEvent *e)
{
	if(isEnabled())startBlinking();
	SSEXTableView::focusInEvent(e);
	emit gainedFocus(this);
}

void SSEXEditor::focusOutEvent(QFocusEvent *e)
{
	stopBlinking();
	SSEXTableView::focusOutEvent(e);
}

void SSEXEditor::setEnabled(bool bEnabled)
{
	SSEXTableView::setEnabled(bEnabled);
	if(bEnabled && hasFocus())startBlinking();
}

void SSEXEditor::updateMaxTextWidth()
{
	m_iMaxTextWidth = 0;
	int line = 0;
	for(SSEXEditorTextLine * l=m_pLines->first();l;l=m_pLines->next())
	{
		if(l->width > m_iMaxTextWidth)
		{
			m_iMaxTextWidth = l->width;
			m_iMaxTextWidthLine = line;
		}
		line++;
	}
}

//////////////////////////////////////////////////////////////////////////////////////////////////
//
// PAINT
//
//////////////////////////////////////////////////////////////////////////////////////////////////


void SSEXEditor::paintCell(QPainter * p,int row)
{
	SSEXEditorTextLine * l = m_pLines->at(row);
	if(!l)return;

	__assert(l->text.length() == l->length);
	
	switch(m_mode)
	{
		case Normal:
			paintCellNormal(p,l,row);
			break;
		case Perl:
			paintCellPerl(p,l,row);
			break;
	}
}

void SSEXEditor::paintBlockMark(QPainter * p,SSEXEditorTextLine * l,int row,const QRect & updateR,QColor * pColor)
{
	if((m_blockMark1.y() <= row) && (m_blockMark2.y() >= row))
	{
		int selectionBeginX = SSEX_EDITOR_BORDER;
		int selectionEndX = SSEX_EDITOR_BORDER + cellWidth();
		if(m_blockMark1.y() == row)
		{
			// the selection begins in this row
			selectionBeginX += getTextWidthWithTabsForCursorPosition(l->text.data(),m_blockMark1.x());
		}
		if(m_blockMark2.y() == row)
		{
			// the selection ends in this row
			selectionEndX = SSEX_EDITOR_BORDER + getTextWidthWithTabsForCursorPosition(l->text.data(),m_blockMark2.x());
		}
		p->fillRect(selectionBeginX,updateR.y(),selectionEndX - selectionBeginX,updateR.height(),*pColor);
	}
}

void SSEXEditor::paintCursor(QPainter * p,SSEXEditorTextLine *l,QColor * pColor)
{
	if(!isEnabled())return;

	if((m_iCursorRow != m_iLastCursorRow) || (m_iCursorPosition != m_iLastCursorPosition))
	{
		m_iLastCursorRow = m_iCursorRow;
		m_iLastCursorPosition = m_iCursorPosition;
//		debug("Cursor pos = %d,%d",m_iCursorRow,m_iCursorPosition);
		emit cursorPositionChanged(this,m_iCursorRow,(m_iCursorPosition < l->length) ? m_iCursorPosition : l->length);
	}
	if(!m_bCursorOn)return;

	int curXPos = m_iCursorPositionInPixels + SSEX_EDITOR_BORDER - 1;

	p->setPen(*pColor);

	int rowLowPos = cellHeight() - 1;

	p->drawLine(curXPos,0,curXPos,rowLowPos);
	p->drawLine(curXPos + 1,0,curXPos + 1,rowLowPos);

	p->drawLine(curXPos - 3,0,curXPos + 4,0);
	p->drawLine(curXPos - 3,rowLowPos,curXPos + 4,rowLowPos);
}


void SSEXEditor::paintSelection(QPainter * p,SSEXEditorTextLine *l,int row,QRect &updateR)
{
	if((m_selection1.y() <= row) && (m_selection2.y() >= row))
	{
		int selectionBeginX = SSEX_EDITOR_BORDER;
		int selectionEndX = SSEX_EDITOR_BORDER;
		if(m_selection1.y() == row)
		{
			// the selection begins in this row
			selectionBeginX += getTextWidthWithTabsForCursorPosition(l->text.data(),m_selection1.x());
		}
		if(m_selection2.y() == row)
		{
			// the selection ends in this row
			selectionEndX += getTextWidthWithTabsForCursorPosition(l->text.data(),m_selection2.x());
		} else {
			// selection ends below
			selectionEndX += getTextWidthWithTabsForCursorPosition(l->text.data(),l->length);
		}
		p->setRasterOp(NotROP);
		p->fillRect(selectionBeginX,updateR.y(),selectionEndX - selectionBeginX,updateR.height(),Qt::black);
		p->setRasterOp(CopyROP);
	}
}


void SSEXEditor::paintCellNormal(QPainter * p,SSEXEditorTextLine * l,int row)
{
	QRect updateR = cellUpdateRect();

	QPainter paint(m_pMemBuffer);

	paint.fillRect(updateR,isEnabled() ? g_pSSEXOptions->clrTxtBackground : g_pSSEXOptions->clrTxtDisabledBackground);
	if(m_bHasBlockMark)paintBlockMark(&paint,l,row,updateR,&(g_pSSEXOptions->clrTxtMarkBackground));

	paint.setPen(g_pSSEXOptions->clrTxtNormalText);
	paint.setFont(p->font());

	const char * c = l->text.data();
	int curXPos = SSEX_EDITOR_BORDER;
	int lastTabPos = SSEX_EDITOR_BORDER;

	while(*c)
	{
		if(*c == '\t')
		{
			// only a tab
			while(lastTabPos <= curXPos)lastTabPos += m_iTabsNumPixels;
			curXPos = lastTabPos;
			c++;
		} else {
			// text block
			const char * begin = c;
			int blockWidth = 0;
			while(*c && *c != '\t'){
				blockWidth += m_iCharWidth[(unsigned char)*c];
				c++;
			}
			paint.drawText(curXPos,m_iFontAscent,begin,c - begin);
			curXPos += blockWidth;
		}
	}

	if(m_iCursorRow == row)paintCursor(&paint,l,&(g_pSSEXOptions->clrTxtCursor));
	if(m_bHasSelection)paintSelection(&paint,l,row,updateR);

	int col_xPos,row_yPos;

	p->worldMatrix().map(0,0,&col_xPos,&row_yPos);

	bitBlt(p->device(),col_xPos + updateR.x(),row_yPos + updateR.y(),m_pMemBuffer,updateR.x(),updateR.y(),updateR.width(),updateR.height());

}


bool cmp(const char *token1,const char *token2,int len1,int len2)
{
	if(len1 != len2)return false;
	return (strncmp(token1,token2,len2) == 0);
}



bool cmpnocase(const char *token1,const char *token2,int len1,int len2)
{
	if(len1 != len2)return false;
	return (strncasecmp(token1,token2,len2) == 0);
}

bool cmpnolen(const char *token1,const char *token2,int len)
{
	return (strncmp(token1,token2,len) == 0);
}



void SSEXEditor::paintCellPerl(QPainter * p,SSEXEditorTextLine * l,int row)
{
	QPainter paint(m_pMemBuffer);

	paint.setFont(font());


	QRect updateR = cellUpdateRect();
	paint.fillRect(updateR,isEnabled() ? g_pSSEXOptions->clrPerlBackground : g_pSSEXOptions->clrPerlDisabledBackground);
	if(m_bHasBlockMark)paintBlockMark(&paint,l,row,updateR,&(g_pSSEXOptions->clrPerlMarkBackground));

	const char * c = l->text.data();
	int curXPos = SSEX_EDITOR_BORDER;
	int lastTabPos = SSEX_EDITOR_BORDER;

	const char * begin;
	QColor * pClr;
	int blockWidth;

	bool bInLineComment = false;

	unsigned int bGotEscape = 0;

	while(*c)
	{

		if(bGotEscape > 0)bGotEscape--;
		switch(*c)
		{
			case '\t':
				// a tab character
				while(lastTabPos <= curXPos)lastTabPos += m_iTabsNumPixels;
				curXPos = lastTabPos;
				paint.setPen(QPen(g_pSSEXOptions->clrPerlTab,0,DashLine));
				paint.drawLine(curXPos - m_iTabsNumPixels,0,curXPos - m_iTabsNumPixels,m_iFontAscent);
				c++;
				break;
			case ' ':
				// a set of spaces
				begin = c;
				while(*c == ' ')c++;
				curXPos += (m_iCharWidth[(unsigned char)' '] * (c - begin));
				break;
			default:
				begin      = c;
				blockWidth = 0;
				{
					//
					// EXTRACT TOKEN ####################################################################################
					//
					if(bInLineComment)
					{
						// We are in a line comment...up to the end of the line!
						while((*c) && (*c != ' ') && (*c != '\t'))
						{
							blockWidth += m_iCharWidth[(unsigned char)(*c)];
							c++;
						}
						pClr = &(g_pSSEXOptions->clrPerlComment);

					} else if(	((*c >= 'a') && (*c <= 'z')) ||
								((*c >= 'A') && (*c <= 'Z')) ||
								(*c == '_'))
					{
						// A text token
						blockWidth += m_iCharWidth[(unsigned char)(*c)];
						c++;
						while(	((*c >= 'a') && (*c <= 'z')) ||
								((*c >= 'A') && (*c <= 'Z')) ||
								((*c >= '0') && (*c <= '9')) ||
								(*c == '_'))
						{
							blockWidth += m_iCharWidth[(unsigned char)(*c)];
							c++;
						}
						//pClr = cppModeGetTokenColor(begin,c - begin);
						pClr = &(g_pSSEXOptions->clrPerlNormalText);
					} else if(*c == '#')
					{
						blockWidth += m_iCharWidth[(unsigned char)(*c)];
						bInLineComment = true;
						c++;
						pClr = &(g_pSSEXOptions->clrPerlComment);
					} else if((*c == '$') || (*c == '%') || (*c == '@'))
					{
						// A variable
						blockWidth += m_iCharWidth[(unsigned char)(*c)];
						c++;
						if(bGotEscape)
						{
							pClr = &(g_pSSEXOptions->clrPerlNormalText);
						} else {
							while(	((*c >= 'a') && (*c <= 'z')) ||
									((*c >= 'A') && (*c <= 'Z')) ||
									((*c >= '0') && (*c <= '9')) ||
									(*c == '_'))
							{
								blockWidth += m_iCharWidth[(unsigned char)(*c)];
								c++;
							}
							pClr = &(g_pSSEXOptions->clrPerlVariable);
						}
					} else if(*c == '\\')
					{
						// escape
						bGotEscape = 2;
						blockWidth += m_iCharWidth[(unsigned char)(*c)];
						c++;
						pClr = &(g_pSSEXOptions->clrPerlOperator);
					} else if( (*c == '{') || (*c == '}') )
					{
						blockWidth += m_iCharWidth[(unsigned char)(*c)];
						c++;
						pClr = &(g_pSSEXOptions->clrPerlBrace);
					} else if((*c == '/') || (*c == '(') || (*c == ')') || (*c == '[') || (*c == ']') ||
								(*c == '"') || (*c == '\'') || (*c == '+') || (*c == '-') || (*c == '*') ||
								(*c == '>') || (*c == '<') || (*c == '~') || (*c == '?') || (*c == '=') ||
								(*c == '!') || (*c == '&') || (*c == '|') || (*c == '.') || (*c == ':') ||
								(*c == ';') || (*c == '^') || (*c == ',') )
					{
						blockWidth += m_iCharWidth[(unsigned char)(*c)];
						c++;
						pClr = bGotEscape ? &(g_pSSEXOptions->clrPerlNormalText) : &(g_pSSEXOptions->clrPerlOperator);
					} else {
						while(	(*c) && (*c != ' ') && (*c != '\t') &&
								(*c != '{') && (*c != '}') && (*c != '/') &&
								(*c != '(') && (*c != ')') && (*c != '[') &&
								(*c != ']') && (*c != '"') && (*c != '\'') &&
								(*c != '+') && (*c != '-') && (*c != '*') &&
								(*c != '!') && (*c != '<') && (*c != '.') &&
								(*c != '>') && (*c != '^') && (*c != '%') &&
								(*c != '&') && (*c != '|') && (*c != '=') &&
								(*c != ',') && (*c != ';') && (*c != ':') &&
								(*c != '~') && (*c != '?') && (*c != '#') &&
								(*c != '\\') && (*c != '$') && (*c != '@'))
						{
							blockWidth += m_iCharWidth[(unsigned char)(*c)];
							c++;
						}
						pClr = &(g_pSSEXOptions->clrPerlNormalText);
					}
					//
					// EXTRACT TOKEN ####################################################################################
					//
				}
				paint.setPen(*pClr);
				paint.drawText(curXPos,m_iFontAscent,begin,c-begin);
				curXPos += blockWidth;
				break;
		}
	}

	if(m_iCursorRow == row)paintCursor(&paint,l,&(g_pSSEXOptions->clrPerlCursor));
	if(m_bHasSelection)paintSelection(&paint,l,row,updateR);

	int col_xPos,row_yPos;

	p->worldMatrix().map(0,0,&col_xPos,&row_yPos);

	bitBlt(p->device(),col_xPos + updateR.x(),row_yPos + updateR.y(),m_pMemBuffer,updateR.x(),updateR.y(),updateR.width(),updateR.height());
}





inline void SSEXEditor::recalcCursorPosition(SSEXEditorTextLine *l)
{
	m_iCursorPositionInPixels = getTextWidthWithTabsForCursorPosition(l->text.data(),m_iCursorPosition);
}

int SSEXEditor::cursorCol() const
{
	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	if(!l)return 0;
	return (m_iCursorPosition < l->length) ? m_iCursorPosition : l->length;
}

/////////////////////////////////////////////////////////////////////////////////////////////////
//
// SELECTION HANDLING HELPERS
//
/////////////////////////////////////////////////////////////////////////////////////////////////

void SSEXEditor::ensureSelectionCoherent()
{
	if((m_selection2.y() < m_selection1.y()) || ((m_selection1.y() == m_selection2.y()) && (m_selection2.x() < m_selection1.x())))
	{
		QPoint aux = m_selection1;
		m_selection1 = m_selection2;
		m_selection2 = aux;
	}
}

void SSEXEditor::setSelectionCoords(int x1,int row1,int x2,int row2)
{
	setHasSelection(true);
	m_selection1 = QPoint(x1,row1);
	m_selection2 = QPoint(x2,row2);
}

void SSEXEditor::clearSelection(bool bUpdate)
{
	setHasSelection(false);
	if(bUpdate)update();	
}

void SSEXEditor::selectionCursorMovement(const QPoint &oldCursorPos,const QPoint &newCursorPos)
{
	if(m_bHasSelection)
	{
		if(m_selection1 == oldCursorPos)m_selection1 = newCursorPos;
		else if(m_selection2 == oldCursorPos)m_selection2 = newCursorPos;
		else setSelectionCoords(oldCursorPos.x(),oldCursorPos.y(),newCursorPos.x(),newCursorPos.y());
	} else setSelectionCoords(oldCursorPos.x(),oldCursorPos.y(),newCursorPos.x(),newCursorPos.y());

	ensureSelectionCoherent();
}
/////////////////////////////////////////////////////////////////////////////////////////////////
//
// CURSOR MOVEMENT
//
/////////////////////////////////////////////////////////////////////////////////////////////////

void SSEXEditor::cursorUp(bool bSelect)
{

	if(m_iCursorRow > 0)
	{
		if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(m_iCursorPosition,m_iCursorRow - 1));
		else if(m_bHasSelection)clearSelection(true);

		m_iCursorRow--;
		recalcCursorPosition(m_pLines->at(m_iCursorRow));
		ensureCursorVisible();
		if(bSelect)update();
		else {
			updateCell(m_iCursorRow + 1);
			updateCell(m_iCursorRow);
		}
	}
}

void SSEXEditor::cursorDown(bool bSelect)
{
	if(m_iCursorRow < (m_pLines->count() - 1))
	{
		if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(m_iCursorPosition,m_iCursorRow + 1));
		else if(m_bHasSelection)clearSelection(true);

		m_iCursorRow++;

		recalcCursorPosition(m_pLines->at(m_iCursorRow));
		ensureCursorVisible();
		if(bSelect)update();
		else {
			updateCell(m_iCursorRow - 1);
			updateCell(m_iCursorRow);
		}
	}
}

void SSEXEditor::cursorPageUp(bool bSelect)
{
	if(m_iCursorRow > 0)
	{
		int oldCursorRow = m_iCursorRow;
		m_iCursorRow -=(viewHeight() / cellHeight());
		if(m_iCursorRow < 0)m_iCursorRow = 0;

		if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,oldCursorRow),QPoint(m_iCursorPosition,m_iCursorRow));
		else if(m_bHasSelection)clearSelection(true);

		recalcCursorPosition(m_pLines->at(m_iCursorRow));
		ensureCursorVisible();
		update();
	}
}

void SSEXEditor::cursorPageDown(bool bSelect)
{
	if(m_iCursorRow < (m_pLines->count() - 1))
	{
		int oldCursorRow = m_iCursorRow;
		m_iCursorRow += (viewHeight() / cellHeight());
		if(m_iCursorRow >= m_pLines->count())m_iCursorRow = m_pLines->count() - 1;

		if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,oldCursorRow),QPoint(m_iCursorPosition,m_iCursorRow));
		else if(m_bHasSelection)clearSelection(true);


		recalcCursorPosition(m_pLines->at(m_iCursorRow));
		ensureCursorVisible();
		update();
	}
}

void SSEXEditor::cursorLeft(bool bSelect)
{	
	SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
	if(m_iCursorPosition > l->length)m_iCursorPosition = l->length;
	if(m_iCursorPosition > 0)
	{
		if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(m_iCursorPosition - 1,m_iCursorRow));
		else if(m_bHasSelection)clearSelection(true);

		m_iCursorPosition--;
		recalcCursorPosition(l);
		ensureCursorVisible();
		if(bSelect)update();
		else {
			updateCell(m_iCursorRow);
		}
	} else {
		if(m_iCursorRow > 0)
		{
			QPoint old(m_iCursorPosition,m_iCursorRow);

			m_iCursorRow--;
			l = m_pLines->at(m_iCursorRow);
			m_iCursorPosition = l->length;

			if(bSelect)selectionCursorMovement(old,QPoint(m_iCursorPosition,m_iCursorRow));
			else if(m_bHasSelection)clearSelection(true);

			recalcCursorPosition(l);
			ensureCursorVisible();
			if(bSelect)update();
			else {
				updateCell(m_iCursorRow + 1);
				updateCell(m_iCursorRow);
			}
		}
	}
}

void SSEXEditor::cursorRight(bool bSelect)
{
	SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
	if(l->length > m_iCursorPosition)
	{
		if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(m_iCursorPosition + 1,m_iCursorRow));
		else if(m_bHasSelection)clearSelection(true);

		m_iCursorPosition++;
		recalcCursorPosition(l);
		ensureCursorVisible();
		if(bSelect)update();
		else updateCell(m_iCursorRow);
	} else {
		if(m_iCursorRow < (m_pLines->count() - 1))
		{
			if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(0,m_iCursorRow));
			else if(m_bHasSelection)clearSelection(true);

			m_iCursorRow++;
			m_iCursorPosition = 0;
			m_iCursorPositionInPixels = 0;
			ensureCursorVisible();
			if(bSelect)update();
			else {
				updateCell(m_iCursorRow - 1);
				updateCell(m_iCursorRow);
			}
		}
	}
}

void SSEXEditor::cursorHome(bool bSelect)
{
	if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(0,m_iCursorRow));
	else if(m_bHasSelection)clearSelection(true);

	SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
	if(m_iCursorPosition > l->length)m_iCursorPosition = l->length;
	m_iCursorPosition = 0;
	recalcCursorPosition(l);
	ensureCursorVisible();
	if(bSelect)update();
	else updateCell(m_iCursorRow);

}

void SSEXEditor::cursorEnd(bool bSelect)
{	
	SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
	if(bSelect)selectionCursorMovement(QPoint(m_iCursorPosition,m_iCursorRow),QPoint(l->length,m_iCursorRow));
	else if(m_bHasSelection)clearSelection(true);
	if(m_iCursorPosition > l->length)m_iCursorPosition = l->length;
	m_iCursorPosition = l->length;
	recalcCursorPosition(l);
	ensureCursorVisible();
	if(bSelect)update();
	else updateCell(m_iCursorRow);
}

/////////////////////////////////////////////////////////////////////////////////////////////////
//
// VIEW ADJUSTEMENTS
//
/////////////////////////////////////////////////////////////////////////////////////////////////

void SSEXEditor::ensureCursorVisible()
{
	if(topCell() > m_iCursorRow)
	{
		setTopCell(m_iCursorRow);
	} else if(topCell() == m_iCursorRow){
		// We want it to be totally visible
//		debug("TOP CELL IS HERE");
		setYOffset(topCell() * cellHeight());
	} else {
		if(lastRowVisible() <= m_iCursorRow)setBottomCell(m_iCursorRow);
	}

	if((m_iCursorPositionInPixels) < xOffset())setXOffset(m_iCursorPositionInPixels);
	else  if((m_iCursorPositionInPixels + SSEX_EDITOR_BORDER + 20) > (viewWidth() + xOffset()))
			setXOffset((m_iCursorPositionInPixels + SSEX_EDITOR_BORDER + 20) - viewWidth());
}

void SSEXEditor::setBottomCell(int row)
{
//	int row_yPos = row * cellHeight();

	int nCellsVisible = viewHeight() / cellHeight();
	int topRow = (row - nCellsVisible) + 1;
	setTopCell(topRow);

//	setYOffset(row_yPos - (viewHeight() - cellHeight()));
}

/////////////////////////////////////////////////////////////////////////////////////////////////
//
// TEXT OPERATIONS
//
/////////////////////////////////////////////////////////////////////////////////////////////////

void SSEXEditor::indentSelected()
{
	if(!m_bHasSelection)return;
	if(m_bHasBlockMark)clearBlockMark(false);


	if(m_selection1.x() > 0)m_selection1.setX(0);

	QCString tmp = selectedText();

	int lastIdx = 0;

	int idx;

	while((idx = tmp.find('\n',lastIdx)) != -1)
	{
		tmp.insert(lastIdx,"\t");
		lastIdx = idx + 2;
	}

	tmp.insert(lastIdx,"\t");

	QPoint sel1 = m_selection1;
	QPoint sel2 = m_selection2;

	killSelection(false);
	insertText(tmp);

	setSelectionCoords(0,sel1.y(),sel2.x() + 1,sel2.y());

	update();
}

void SSEXEditor::unindentSelected()
{
//#warning "This should have an UNIQUE undo...not in two steps! (same for indentSelected())"
	if(!m_bHasSelection)return;
	if(m_bHasBlockMark)clearBlockMark(false);

	if(m_selection1.x() > 0)m_selection1.setX(0);

	QCString tmp = selectedText();

	if(tmp.isNull())tmp = "";

	int lastIdx = 0;
	int idx;

	while((idx = tmp.find('\n',lastIdx)) != -1)
	{
		if(*(tmp.data() + lastIdx) == '\t')
		{
			tmp.remove(lastIdx,1);
			lastIdx = idx;
		} else {
			lastIdx = idx + 1;
		}
	}

	bool bLast = false;

	if(*(tmp.data() + lastIdx) == '\t')
	{
		bLast = true;
		tmp.remove(lastIdx,1);
	}

	QPoint sel1 = m_selection1;
	QPoint sel2 = m_selection2;

	killSelection(false);
	insertText(tmp);

	setSelectionCoords(0,sel1.y(),(sel2.x() > 0 && bLast) ? sel2.x() - 1: sel2.x(),sel2.y());

	update();
}

void SSEXEditor::removeCR()
{
	for(SSEXEditorTextLine * l = m_pLines->first();l;l = m_pLines->next())
	{
		int len = l->text.length() - 1;
		if(len >= 0)
		{
			const char * t = l->text.data();
			while(*(t + len) == '\r')
			{
				l->text.remove(len,1);
				l->length--;
				l->width = getTextWidthWithTabs(l->text.data());

				if(m_iMaxTextWidthLine == m_iCursorRow)
				{
					updateMaxTextWidth();
					updateCellSize();
				}
			}
		}
	}
	recalcCursorPosition(m_pLines->at(m_iCursorRow));
	ensureCursorVisible();
	update();
	setModified(true);
}

void SSEXEditor::indent()
{
	if(m_bHasSelection)
	{
		indentSelected();
		return;
	}

	if(m_bHasBlockMark)clearBlockMark(false);

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	addUndo(new SSEXUndo("indent line",m_iCursorRow,0,m_iCursorRow,1));
	l->text.insert(0,'\t');
	l->length++;
	l->width  = getTextWidthWithTabs(l->text.data());
	if(m_iMaxTextWidthLine == m_iCursorRow)
	{
		m_iMaxTextWidth = l->width;
		updateCellSize();
	} else {
		if(m_iMaxTextWidth < l->width)
		{
			m_iMaxTextWidth = l->width;
			m_iMaxTextWidthLine = m_iCursorRow;
			updateCellSize();
		}
	}

	if(m_iCursorRow < m_pLines->count() - 1)m_iCursorRow++;
	recalcCursorPosition(m_pLines->at(m_iCursorRow));

	ensureCursorVisible();
	updateCell(m_iCursorRow);
	updateCell(m_iCursorRow - 1);

	setModified(true);
}

void SSEXEditor::unindent()
{
	if(m_bHasSelection)
	{
		unindentSelected();
		return;
	}

	if(m_bHasBlockMark)clearBlockMark(true);

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	if(*(l->text.data()) == '\t')
	{
		addUndo(new SSEXUndo("unindent line",m_iCursorRow,0,"\t"));
		l->text.remove(0,1);
		l->length--;
		l->width = getTextWidthWithTabs(l->text.data());
		if(m_iMaxTextWidthLine == m_iCursorRow)
		{
			updateMaxTextWidth();
			updateCellSize();
		}
	} else l=0;
	if(m_iCursorRow < m_pLines->count() - 1)m_iCursorRow++;
	recalcCursorPosition(m_pLines->at(m_iCursorRow));
	ensureCursorVisible();
	updateCell(m_iCursorRow);
	updateCell(m_iCursorRow - 1);
	if(l)setModified(true);
}

bool SSEXEditor::curTextLine(QCString &buf)
{
	SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
	if(l)buf = l->text;
	return (l != 0);
}

void SSEXEditor::commentOut(bool bAlternative)
{
	if(m_bHasBlockMark)clearBlockMark(true);
	if(m_bHasSelection)clearSelection(true);

	if(m_mode == Perl)
	{
		SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
		addUndo(new SSEXUndo("add shell comment",m_iCursorRow,l->length,m_iCursorRow,l->length + 1));
		l->text.insert(0,"#");
		l->length += 1;
		l->width  = getTextWidthWithTabs(l->text.data());
		if(m_iMaxTextWidthLine == m_iCursorRow){
			m_iMaxTextWidth = l->width;
			updateCellSize();
		} else {
			if(m_iMaxTextWidth < l->width)
			{
				m_iMaxTextWidth = l->width;
				m_iMaxTextWidthLine = m_iCursorRow;
				updateCellSize();
			}
		}

		if(m_iCursorRow < m_pLines->count() - 1)m_iCursorRow++;
		recalcCursorPosition(m_pLines->at(m_iCursorRow));

		ensureCursorVisible();
		updateCell(m_iCursorRow);
		updateCell(m_iCursorRow - 1);
		setModified(true);
	}
}

void SSEXEditor::removeComment()
{
	if(m_bHasBlockMark)clearBlockMark(true);
	if(m_bHasSelection)clearSelection(true);

	if(m_mode == Perl)
	{

		SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

		if(strncmp(l->text.data(),"#",1) == 0)
		{
			addUndo(new SSEXUndo("remove shell comment",m_iCursorRow,0,"#"));
			l->text.remove(0,1);
			l->length -= 1;
			l->width = getTextWidthWithTabs(l->text.data());
			if(m_iMaxTextWidthLine == m_iCursorRow){
				updateMaxTextWidth();
				updateCellSize();
			}
		} else l = 0;

		if(m_iCursorRow < m_pLines->count() - 1)m_iCursorRow++;
		recalcCursorPosition(m_pLines->at(m_iCursorRow));
		ensureCursorVisible();
		updateCell(m_iCursorRow);
		updateCell(m_iCursorRow - 1);
		if(l)setModified(true);
	}
}

void SSEXEditor::insertChar(char c)
{
#ifdef __DEBUG
	if((((unsigned char)c) < 32) && (((unsigned char)c) != ((unsigned char)'\t')))debug("INSERTING A STRANGE CHAR! (%c:%d)",c,c);
#endif

	if(m_bHasBlockMark)clearBlockMark(true);
	if(m_bHasSelection)killSelection(true); // trigger the total update too

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	char deletedChar = 0;
	if(m_iCursorPosition <= l->length)
	{
		if(m_bOverwrite && (m_iCursorPosition < l->length))
		{
			deletedChar = *(l->text.data() + m_iCursorPosition);
			addUndo(new SSEXUndo("kill char",m_iCursorRow,m_iCursorPosition,l->text.mid(m_iCursorPosition,1)));
			l->text.remove(m_iCursorPosition,1);
		} else l->length++;
		addUndo(new SSEXUndo("insert char",m_iCursorRow,m_iCursorPosition,m_iCursorRow,m_iCursorPosition + 1));
		l->text.insert(m_iCursorPosition,c);
		m_iCursorPosition++;
	} else {
		addUndo(new SSEXUndo("insert char",m_iCursorRow,l->length,m_iCursorRow,l->length + 1));
		l->text.insert(l->length,c);
		l->length++;
		m_iCursorPosition = l->length;
	}
	l->width  = getTextWidthWithTabs(l->text.data());
	recalcCursorPosition(l);
	if(m_iMaxTextWidthLine == m_iCursorRow)
	{
		if(m_bOverwrite)updateMaxTextWidth();
		else m_iMaxTextWidth = l->width;
		updateCellSize();
	} else {
		if(m_iMaxTextWidth < l->width)
		{
			m_iMaxTextWidth = l->width;
			m_iMaxTextWidthLine = m_iCursorRow;
			updateCellSize();
		}
	}

	__assert(l->text.length() == l->length);

	ensureCursorVisible();
	updateCell(m_iCursorRow);

	setModified(true);
}

void SSEXEditor::newLine()
{
	if(m_bHasSelection)clearSelection(false);
	if(m_bHasBlockMark)clearBlockMark(false);


	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	int pos = ((m_iCursorPosition <= l->length) ? m_iCursorPosition : l->length);

	int oldCursorRow = m_iCursorRow;

	SSEXEditorTextLine *n = new SSEXEditorTextLine;
	n->text = (l->text.data() + pos);
	n->length = n->text.length();
	l->length -= n->length;
	l->text.truncate(l->length);
	l->width  = getTextWidthWithTabs(l->text.data());
	n->width  = getTextWidthWithTabs(n->text.data());
	m_pLines->insert(m_iCursorRow + 1,n);
	setNumRows(m_pLines->count());
	if(m_iMaxTextWidthLine == m_iCursorRow){
		updateMaxTextWidth();
		updateCellSize();
	} else {
		// MaxTextWidthLine is scrolled down!
		if(m_iMaxTextWidthLine > m_iCursorRow)m_iMaxTextWidthLine++;
	}
	m_iCursorRow++;
	m_iCursorPosition = 0;

	addUndo(new SSEXUndo("new line",oldCursorRow,pos,m_iCursorRow,m_iCursorPosition));

	recalcCursorPosition(l);

	ensureCursorVisible();
	update();
	setModified(true);
}

void SSEXEditor::backspace()
{
	if(m_bHasBlockMark)clearBlockMark(true);

	if(m_bHasSelection)
	{
		killSelection(true);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	int pos = (m_iCursorPosition <= l->length) ? m_iCursorPosition : l->length;

	if(pos == 0){
		// join lines
		if(m_iCursorRow > 0){
			m_iCursorRow--;
			SSEXEditorTextLine *prev = m_pLines->at(m_iCursorRow);
			prev->text.append(l->text);
			m_iCursorPosition = prev->length;
			recalcCursorPosition(prev);
			prev->length += l->length;
			prev->width = getTextWidthWithTabs(prev->text.data());
			m_pLines->removeRef(l);
			setNumRows(m_pLines->count());

			if((m_iMaxTextWidthLine == m_iCursorRow)||(m_iMaxTextWidthLine == (m_iCursorRow + 1))){
				m_iMaxTextWidth = prev->width;
				m_iMaxTextWidthLine = m_iCursorRow;
				updateCellSize();
			} else {
				if(m_iMaxTextWidth < prev->width)
				{
					m_iMaxTextWidth = prev->width;
					m_iMaxTextWidthLine = m_iCursorRow;
					updateCellSize();
				} else {
					// The maxtextwidthline is scrolled up!
					if(m_iMaxTextWidthLine > (m_iCursorRow + 1))m_iMaxTextWidthLine--;
				}
			}

			addUndo(new SSEXUndo("join lines",m_iCursorRow,m_iCursorPosition,"\n"));

			update(); // we could calc the right rect here
		}
	} else {
		char c = *(l->text.data() + pos - 1);

		QCString undoText = l->text.mid(pos - 1,1);

		l->text.remove(pos - 1,1);
		l->length--;
		l->width = getTextWidthWithTabs(l->text.data());
		m_iCursorPosition = pos - 1;
		recalcCursorPosition(l);
		if(m_iMaxTextWidthLine == m_iCursorRow)
		{
			updateMaxTextWidth();
			updateCellSize();
		}
		pos--;
		addUndo(new SSEXUndo("delete char",m_iCursorRow,m_iCursorPosition,undoText));
		ensureCursorVisible();
		updateCell(m_iCursorRow);
	}
	setModified(true);
}

void SSEXEditor::goToLineNum(int lineNum)
{
	if(lineNum < 0)return;
	SSEXEditorTextLine * l = m_pLines->at(lineNum);
	if(l)
	{
		clearSelection();
		m_iCursorRow = lineNum;
		m_iCursorPosition = 0;
		recalcCursorPosition(l);
		ensureCursorVisible();
	}
}

void SSEXEditor::goToLine()
{
	int lineNum = QInputDialog::getInteger("Go to line","Insert the line number");
	lineNum--;
	goToLineNum(lineNum);
}

void SSEXEditor::goToBeginning()
{
	goToLineNum(0);
}

void SSEXEditor::goToEnd()
{
	if(m_pLines->count() > 0)goToLineNum(m_pLines->count() - 1);
	cursorEnd(false);
}

QCString SSEXEditor::getTokenBeforeCursor()
{
	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	if(m_iCursorPosition > l->length)m_iCursorPosition = l->length;
	if(m_iCursorPosition < 1)
	{
		QCString ret;
		return ret;
	}
	char * end = (l->text.data() + m_iCursorPosition);
	char * aux = end - 1;
	char * beg = l->text.data();
	while(aux > beg)
	{
		if(!(isalnum(*aux) || (*aux == '_')))
		{
			aux++;
			break;
		}
		aux--;
	}
	if(aux == beg)if(!(isalnum(*aux) || (*aux == '_')))aux++;
	if(end == aux)
	{
		QCString ret;
		return ret;
	}
	QCString token(aux,(end - aux) + 1);
	return token;
}

void SSEXEditor::quickCompletion()
{
	QCString token = getTokenBeforeCursor();
	if(token.isEmpty())return;

	SSEXCompletion c;
	int limit = m_iCursorRow - 50;
	if(limit < 0)limit = 0;
	int i = m_iCursorRow;
	QCString tmp = m_pLines->at(i)->text;
	tmp.resize(m_iCursorPosition + 1);
	for(i--;i > limit;i--)c.tokenize(m_pLines->at(i)->text.data());
	QCString cmp = c.getCompletion(token.data());
	insertText(cmp);
}


void SSEXEditor::del()
{
	if(m_bHasBlockMark)clearBlockMark(true);

	if(m_bHasSelection)
	{
		killSelection(true);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
	if(m_iCursorPosition > l->length)m_iCursorPosition = l->length;

	if(m_iCursorPosition == l->length){
		// join lines

		if(m_iCursorRow < (m_pLines->count() - 1))
		{
			SSEXEditorTextLine *next = m_pLines->at(m_iCursorRow + 1);
			l->text.append(next->text);
			l->length += next->length;
			l->width = getTextWidthWithTabs(l->text.data());
			m_pLines->removeRef(next);

			setNumRows(m_pLines->count());
			if((m_iMaxTextWidthLine == m_iCursorRow)||(m_iMaxTextWidthLine == (m_iCursorRow + 1))){
				m_iMaxTextWidth = l->width;
				m_iMaxTextWidthLine = m_iCursorRow;
				updateCellSize();
			} else {
				if(m_iMaxTextWidth < l->width)
				{
					m_iMaxTextWidth = l->width;
					m_iMaxTextWidthLine = m_iCursorRow;
					updateCellSize();
				}
			}

			addUndo(new SSEXUndo("join lines",m_iCursorRow,m_iCursorPosition,"\n"));

			update(); // we could calc the right rect here
		}
	} else {

		char c = *(l->text.data() + m_iCursorPosition);

		addUndo(new SSEXUndo("kill char",m_iCursorRow,m_iCursorPosition,l->text.mid(m_iCursorPosition,1)));

		l->text.remove(m_iCursorPosition,1);
		l->length--;
		l->width = getTextWidthWithTabs(l->text.data());
		if(m_iMaxTextWidthLine == m_iCursorRow)
		{
			updateMaxTextWidth();
			updateCellSize();
		}
		ensureCursorVisible();
		updateCell(m_iCursorRow);
	}
	setModified(true);
}

void SSEXEditor::copy()
{
	if(m_bHasSelection)QApplication::clipboard()->setText(selectedText());
}

void SSEXEditor::paste()
{
	QCString cText = QApplication::clipboard()->text().ascii();
	if(!(cText.isNull() || cText.isEmpty()))insertText(cText);
}

void SSEXEditor::cut()
{
	if(!m_bHasSelection)return;
	copy();
	if(m_bHasBlockMark)clearBlockMark(false);
	killSelection(true);
}


void SSEXEditor::addUndo(SSEXUndo * und)
{
	m_pUndoList->append(und);
	while(m_pUndoList->count() > g_pSSEXOptions->uUndoDepth)m_pUndoList->removeFirst();
}

void SSEXEditor::clearUndoList()
{
	delete m_pUndoList;
	m_pUndoList = new KviPtrList<SSEXUndo>;
	m_pUndoList->setAutoDelete(true);
}

void SSEXEditor::undo()
{
	SSEXUndo * u = m_pUndoList->last();
	if(!u)return;
	if(u->m_type == SSEXUndo::Insert)
	{
		// Insert
		if(m_bHasSelection)clearSelection(true);
		m_iCursorRow = u->m_iRow1;
		m_iCursorPosition = u->m_iCol1;
		insertText(u->m_szText,true,true,false);
	} else {
		// Remove
		if(m_bHasSelection)clearSelection(true);
		setSelectionCoords(u->m_iCol1,u->m_iRow1,u->m_iCol2,u->m_iRow2);
		killSelection(true,true,false);
	}
	m_pUndoList->removeLast();
}

void SSEXEditor::killSelection(bool bUpdate,bool bRecalcWidths,bool bCreateUndo)
{
	if(!m_bHasSelection)return;
	SSEXEditorTextLine * first = m_pLines->at(m_selection1.y());
	if(!first)return;
	// Single line selection
	QCString undoText;
	if(m_selection1.y() == m_selection2.y())
	{
		if(bCreateUndo)undoText = first->text.mid(m_selection1.x(),m_selection2.x() - m_selection1.x());
		first->text.remove(m_selection1.x(),m_selection2.x() - m_selection1.x());
		first->length = first->text.length();
	} else {
		// Multiple lines selection
		if(bCreateUndo)undoText = first->text.mid(m_selection1.x(),first->length - m_selection1.x());
		first->text.remove(m_selection1.x(),first->length - m_selection1.x());
		int idx = m_selection1.y() + 1;
		SSEXEditorTextLine * l = m_pLines->next();
		KviPtrList<SSEXEditorTextLine> list;
		list.setAutoDelete(false);
		while(l && (idx <= m_selection2.y()))
		{
			if(bCreateUndo)undoText.append("\n");
			if(idx == m_selection2.y())
			{
				if(bCreateUndo)undoText.append(l->text.left(m_selection2.x()));

				l->text.remove(0,m_selection2.x());
				first->text.append(l->text);
				list.append(l);
			} else {
				if(bCreateUndo)undoText.append(l->text);

				list.append(l);
			}
			l = m_pLines->next();
			idx++;
		}
		first->length = first->text.length();
		// Kill the middle lines
		for(SSEXEditorTextLine *line=list.first();line;line=list.next())m_pLines->removeRef(line);
		setNumRows(m_pLines->count());
	}

	first->width = getTextWidthWithTabs(first->text.data());

	setHasSelection(false);
	m_iCursorRow = m_selection1.y();
	m_iCursorPosition = m_selection1.x() < first->length ? m_selection1.x() : first->length;
	
	if(bCreateUndo)addUndo(new SSEXUndo("kill selection",m_iCursorRow,m_iCursorPosition,undoText));


	recalcCursorPosition(first);
	if(bRecalcWidths)
	{
		updateMaxTextWidth();
		updateCellSize();
	}
	if(bUpdate){
		ensureCursorVisible();
		update();
	}
	setModified(true);
}

void SSEXEditor::insertText(QCString &text,bool bUpdate,bool bRecalcWidths,bool bCreateUndo)
{
	if(m_bHasSelection)killSelection(false,false);
	if(m_bHasBlockMark)clearBlockMark(false);

	SSEXEditorTextLine * first = m_pLines->at(m_iCursorRow);
	if(!first)return;



	QCString rightPart = (m_iCursorPosition < first->length) ? (first->text.data() + m_iCursorPosition) : "";
	__assert(first->text.length() == first->length);
	first->text.remove(m_iCursorPosition,first->length - m_iCursorPosition);

	SSEXEditorTextLine * l = first;
	int row = m_iCursorRow;

	int oldCursorRow = m_iCursorRow;
	int oldCursorPosition = m_iCursorPosition;

	QCString theText = text;
	int idx = theText.find('\n');
	while(idx != -1)
	{
		l->text.append(theText.left(idx));
		theText.remove(0,idx + 1);
		l->length = l->text.length();
		l->width = getTextWidthWithTabs(l->text.data());
		row++;
		l = new SSEXEditorTextLine;
		l->text = "";
		l->length = 0;
		l->width = 0;
		m_pLines->insert(row,l);
		idx = theText.find('\n');
	}
	setNumRows(m_pLines->count());

	m_iCursorRow = row;
	m_iCursorPosition = l->text.length() + theText.length(); 

	if(bCreateUndo)addUndo(new SSEXUndo("insert text",oldCursorRow,oldCursorPosition,m_iCursorRow,m_iCursorPosition));

	l->text.append(theText);
	l->text.append(rightPart);
	l->length = l->text.length();
	l->width = getTextWidthWithTabs(l->text.data());


	recalcCursorPosition(l);
	if(bRecalcWidths)
	{
		updateMaxTextWidth();
		updateCellSize();
	}
	if(bUpdate)
	{
		ensureCursorVisible();
		update();
	}
	setModified(true);
}

void SSEXEditor::wheelEvent(QWheelEvent *e)
{
 	if(e->delta () < 0)
	{
 		cursorDown(0);
 	} else {
 		cursorUp(0);
 	}
}

QCString SSEXEditor::selectedText()
{
	QCString ret = "";
	if(!m_bHasSelection)return ret;
	SSEXEditorTextLine * l = m_pLines->at(m_selection1.y());
	if(!l)return ret;
	if(m_selection1.y() == m_selection2.y()){
		return l->text.mid(m_selection1.x(),m_selection2.x() - m_selection1.x());
	} else {
		if(m_selection1.x() < l->length)ret = (l->text.data() + m_selection1.x());
		int idx = m_selection1.y();
		do{
			l = m_pLines->next();
			idx++;
			ret.append("\n");
			if(m_selection2.y() == idx)
			{
				ret.append(l->text.left(m_selection2.x()));
			} else ret.append(l->text);
		} while(l && (idx < m_selection2.y()));
	}
	return ret;	
}

void SSEXEditor::findNext()
{
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull()){
		QMessageBox::warning(this,"Find Next - KVIrc","No text to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	int row = m_iCursorRow;
	int col = m_iCursorPosition;

	while(l){

		if(col < l->length){
			int idx = l->text.find(toFind.data(),col,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());
			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + toFind.length();
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,m_iCursorPosition,row);
				ensureCursorVisible();
				m_bCursorOn = true;
				update();
				setFocus();
				return;
			}
		}
		if(row >= (m_pLines->count() - 1))
		{
			// End reached
			if(QMessageBox::information(this,"Find Next - KVIrc","No matches found.\nContinue from the beginning of the file?",
				QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) != QMessageBox::Yes)return;
			row = 0;

		} else row++;

		col = 0;
		l = m_pLines->at(row);
	}
	// newer here
}

void SSEXEditor::findPrev()
{
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull())
	{
		QMessageBox::warning(this,"Find Previous - KVIrc","No text to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	int row = m_iCursorRow;
	int col = m_iCursorPosition - 1;

	while(l){

		if(col >= (toFind.length() - 1)){
			int idx = l->text.findRev(toFind.data(),col,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());
			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx;
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,idx + toFind.length(),row);
				ensureCursorVisible();
				m_bCursorOn = true;
				update();
				setFocus();
				return;
			}
		}
		if(row <= 0)
		{
			// End reached
			if(QMessageBox::information(this,"Find Previous - KVIrc","No matches found.\nContinue from the end of the file?",
				QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) != QMessageBox::Yes)return;
			row = m_pLines->count() - 1;

		} else row--;

		l = m_pLines->at(row);
		col = l->length;
	}
	// newer here
}

void SSEXEditor::findRegExp(QString &regexp)
{
	int row = 0;
	int col = 0;

	QRegExp rx(regexp,true);

	int len = 1;

	SSEXEditorTextLine *l = m_pLines->at(0);


	while(l){

		if(col < l->length)
		{
#if QT_VERSION >= 0x030100
			int idx = rx.search(QString(l->text.data()),col);
			len = rx.matchedLength();
#else
			int idx = rx.match(l->text,col,&len);
#endif
			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + len;
				recalcCursorPosition(l);
//				setSelectionCoords(idx,row,m_iCursorPosition,row);
				ensureCursorVisible();
				m_bCursorOn = true;
				update();
				setFocus();
				return;
			}
		}
		if(row >= (m_pLines->count() - 1))
		{
			// End reached
			return;

		} else row++;

		col = 0;
		l = m_pLines->at(row);
	}
}

void SSEXEditor::findNextRegExp()
{
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull()){
		QMessageBox::warning(this,"Find Next (Regular Expression) - KVIrc","No regular expression to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	int row = m_iCursorRow;
	int col = m_iCursorPosition;

	QRegExp rx(toFind,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());

	int len = 1;

	while(l){

		if(col < l->length){
#if QT_VERSION >= 0x030100
			int idx = rx.search(QString(l->text.data()),col);
			len = rx.matchedLength();
#else
			int idx = rx.match(l->text,col,&len);
#endif
			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + len;
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,m_iCursorPosition,row);
				ensureCursorVisible();
				m_bCursorOn = true;
				update();
				setFocus();
				return;
			}
		}
		if(row >= (m_pLines->count() - 1))
		{
			// End reached
			if(QMessageBox::information(this,"Find Next (Regular Expression) - KVIrc","No matches found.\nContinue from the beginning of the file?",
				QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) != QMessageBox::Yes)return;
			row = 0;

		} else row++;

		col = 0;
		l = m_pLines->at(row);
	}
	// newer here
}

void SSEXEditor::findPrevRegExp()
{
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull()){
		QMessageBox::warning(this,"Find Previous (Regular Expression) - KVIrc","No regular expression to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	int row = m_iCursorRow;
	int col = m_iCursorPosition - 1;

	QRegExp rx(toFind,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());
	int len = 1;

	while(l){

		if(col >= (toFind.length() - 1)){
			int idx = col;
			bool bFound = false;

			while((idx >= 0) && (!bFound)){
#if QT_VERSION >= 0x030100
				if(rx.search(QString(l->text.data()),idx) != idx)idx--;
				else bFound = true;
#else
				if(rx.match(l->text,idx,&len) != idx)idx--;
				else bFound = true;
#endif
			}

			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx;
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,idx + len,row);
				ensureCursorVisible();
				m_bCursorOn = true;
				update();
				setFocus();
				return;
			}
		}
		if(row <= 0)
		{
			// End reached
			if(QMessageBox::information(this,"Find Previous (Regular Expression) - KVIrc","No matches found.\nContinue from the end of the file?",
				QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) != QMessageBox::Yes)return;
			row = m_pLines->count() - 1;

		} else row--;

		l = m_pLines->at(row);
		col = l->length;
	}
	// newer here
}

void SSEXEditor::replace()
{
	if(!m_bHasSelection)
	{
		QMessageBox::warning(this,"Replace - KVIrc","No text selected.",QMessageBox::Ok,0);
		return;
	}
	QCString toReplace = m_pFindWidget->m_pReplaceStringEdit->text().ascii();
	if(toReplace.isNull())toReplace = "";
	insertText(toReplace);
	setFocus();
}

void SSEXEditor::replaceAndFindNext()
{
	replace();
	findNext();
}

void SSEXEditor::replaceAndFindNextRegExp()
{
	replace();
	findNextRegExp();
}

void SSEXEditor::replaceAllInSelection()
{
	if(!m_bHasSelection)
	{
		QMessageBox::warning(this,"Replace All in Selection - KVIrc","No selection to search in.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	QPoint topLeft = m_selection1;
	QPoint bottomRight = m_selection2;

	clearSelection(false);

	QCString toReplace = m_pFindWidget->m_pReplaceStringEdit->text().ascii();
	if(toReplace.isNull())toReplace = "";
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull())
	{
		QMessageBox::warning(this,"Replace All in Selection - KVIrc","No text to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(topLeft.y());

	int row = topLeft.y();
	int col = topLeft.x();

	bool bFound;
	int count = 0;

	int lineCharsDiff = 0;

	SSEXEditorTextLine * lastLine = 0;

	while(l && (row <= bottomRight.y())){

		bFound = false;

		if(lastLine != l)
		{
			lastLine = l;
			lineCharsDiff = -(l->length);
		}

		if(col < l->length)
		{
			int idx = l->text.find(toFind.data(),col,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());
			if(idx != -1)
			{
				if(row == bottomRight.y())
				{
					if((idx + toFind.length()) > bottomRight.x())
					{
						// finished
						break;
					}
				}
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + toFind.length();
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,m_iCursorPosition,row);
				insertText(toReplace,false,false);
				col = m_iCursorPosition;
				bFound = true;
				count ++;
			} else bFound = false;
		}


		if(!bFound)
		{
			lineCharsDiff += l->length;
			row++;
			col = 0;
			l = m_pLines->at(row);
		}
	}

	setSelectionCoords(topLeft.x(),topLeft.y(),bottomRight.x() + lineCharsDiff,bottomRight.y());

	updateMaxTextWidth();
	updateCellSize();
	m_bCursorOn = true;
	ensureCursorVisible();
	update();

	QCString msg;
	msg.sprintf("Replaced %d occurrences",count);
	emit textMessage(this,msg);
	setFocus();
}

void SSEXEditor::replaceAllInSelectionRegExp()
{
	if(!m_bHasSelection)
	{
		QMessageBox::warning(this,"Replace All in Selection (Regular Expression) - KVIrc","No selection to search in.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	QPoint topLeft = m_selection1;
	QPoint bottomRight = m_selection2;

	clearSelection(false);

	QCString toReplace = m_pFindWidget->m_pReplaceStringEdit->text().ascii();
	if(toReplace.isNull())toReplace = "";
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull()){
		QMessageBox::warning(this,"Replace All in Selection (Regular Expression) - KVIrc","No regular expression to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	QRegExp rx(toFind,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());

	SSEXEditorTextLine *l = m_pLines->at(topLeft.y());

	int row = topLeft.y();
	int col = topLeft.x();

	bool bFound;
	int count = 0;
	int len = 1;

	int lineCharsDiff = 0;

	SSEXEditorTextLine * lastLine = 0;

	while(l && (row <= bottomRight.y())){

		bFound = false;

		if(lastLine != l)
		{
			lastLine = l;
			lineCharsDiff = -(l->length);
		}

		if(col < l->length){
#if QT_VERSION >= 0x030100
			int idx = rx.search(QString(l->text.data()),col);
			len = rx.matchedLength();
#else
			int idx = rx.match(l->text,col,&len);
#endif
			if(idx != -1)
			{
				if(row == bottomRight.y())
				{
					if((idx + len) > bottomRight.x())
					{
						// finished
						break;
					}
				}
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + len;
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,m_iCursorPosition,row);
				insertText(toReplace,false,false);
				col = m_iCursorPosition;
				bFound = true;
				count ++;
			} else bFound = false;
		}


		if(!bFound)
		{
			lineCharsDiff += l->length;
			row++;
			col = 0;
			l = m_pLines->at(row);
		}
	}

	setSelectionCoords(topLeft.x(),topLeft.y(),bottomRight.x() + lineCharsDiff,bottomRight.y());

	updateMaxTextWidth();
	updateCellSize();
	m_bCursorOn = true;
	ensureCursorVisible();
	update();

	QCString msg;
	msg.sprintf("Replaced %d occurrences",count);
	emit textMessage(this,msg);
	setFocus();
}



void SSEXEditor::replaceAll()
{
	if(QMessageBox::warning(this,"Replace All - KVIrc",
		"This may be a dangerous action.<br>" \
		"Really replace all occurrences from<br>" \
		"the cursor position to end of the file?",
		QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) == QMessageBox::No)return;

	QCString toReplace = m_pFindWidget->m_pReplaceStringEdit->text().ascii();
	if(toReplace.isNull())toReplace = "";
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull()){
		QMessageBox::warning(this,"Replace All - KVIrc","No text to find.",QMessageBox::Ok,0);
		return;
	}

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	int row = m_iCursorRow;
	int col = m_iCursorPosition;

	bool bFound;
	int count = 0;

	while(l){

		bFound = false;

		if(col < l->length){
			int idx = l->text.find(toFind.data(),col,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());
			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + toFind.length();
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,m_iCursorPosition,row);
				insertText(toReplace,false,false);
				col = m_iCursorPosition;
				bFound = true;
				count ++;
			} else bFound = false;
		}

		if(!bFound)
		{
			if(row >= (m_pLines->count() - 1))
			{
				// End reached
				updateMaxTextWidth();
				updateCellSize();
				m_bCursorOn = true;
				ensureCursorVisible();
				update();
				QCString msg;
				msg.sprintf("Replaced %d occurrences",count);
				emit textMessage(this,msg);
				if(QMessageBox::information(this,"Replace All - KVIrc","No more occurrences found.\nContinue from the beginning?",
					QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) != QMessageBox::Yes)
				{
					setFocus();
					return;
				}
				row = 0;
				count = 0;
			} else row++;

			col = 0;
			l = m_pLines->at(row);
		}
	}
}

void SSEXEditor::selectAll()
{
	int cnt = m_pLines->count();
	if(cnt > 0)cnt--;
	SSEXEditorTextLine * l = m_pLines->last();
	setSelectionCoords(0,0,l->length,cnt);
	m_bCursorOn = true;
	ensureCursorVisible();
	update();	
}

void SSEXEditor::replaceAllRegExp()
{
	if(QMessageBox::warning(this,"Replace All (Regular Expression) - KVIrc",
		"This may be a dangerous action.<br>" \
		"Really replace all occurrences from<br>" \
		"the cursor position to end of the file?",
		QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) == QMessageBox::No)return;


	QCString toReplace = m_pFindWidget->m_pReplaceStringEdit->text().ascii();
	if(toReplace.isNull())toReplace = "";
	QCString toFind = m_pFindWidget->m_pFindStringEdit->text().ascii();
	if(toFind.isEmpty() || toFind.isNull()){
		QMessageBox::warning(this,"Replace All - KVIrc","No regular expression to find.",QMessageBox::Ok | QMessageBox::Default,0);
		return;
	}

	QRegExp rx(toFind,m_pFindWidget->m_pCaseSensitiveCheckBox->isChecked());

	SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);

	int row = m_iCursorRow;
	int col = m_iCursorPosition;

	bool bFound;
	int len = 1;

	while(l){

		bFound = false;

		if(col < l->length){
#if QT_VERSION >= 0x030100
			int idx = rx.search(QString(l->text.data()),col);
			len = rx.matchedLength();
#else
			int idx = rx.match(l->text,col,&len);
#endif
			if(idx != -1)
			{
				// Found an occurrence
				m_iCursorRow = row;
				m_iCursorPosition = idx + len;
				recalcCursorPosition(l);
				setSelectionCoords(idx,row,m_iCursorPosition,row);
				insertText(toReplace,false,false);
				col = m_iCursorPosition;
				bFound = true;
			} else bFound = false;
		}

		if(!bFound)
		{
			if(row >= (m_pLines->count() - 1))
			{
				// End reached
				updateMaxTextWidth();
				updateCellSize();
				m_bCursorOn = true;
				ensureCursorVisible();
				update();
				if(QMessageBox::information(this,"Replace All (Regular Expression) - KVIrc","No more occurrences found.\nContinue from the beginning?",
					QMessageBox::Yes | QMessageBox::Default,QMessageBox::No | QMessageBox::Escape) != QMessageBox::Yes)
				{
					setFocus();
					return;
				}
				row = 0;

			} else row++;

			col = 0;
			l = m_pLines->at(row);
		}
	}

}

void SSEXEditor::switchMode()
{
	switch(m_mode)
	{
		case Normal: setMode(Perl);      break;
		case Perl:   setMode(Normal); break;
		default:     setMode(Normal);   break;
	}
}

void SSEXEditor::setMode(ColorMode mode)
{
	if(mode != m_mode)
	{
		m_mode = mode;
		emit modeChanged(this);
		initializeCurrentMode();
		update();
	}
}

void SSEXEditor::recordKeystrokes()
{
	if(m_bRecordingKeystrokes)
	{
		m_bRecordingKeystrokes = false;
		emit recordingKeystrokes(false);
		return;
	}
	m_pKeystrokes->clear();
	m_bRecordingKeystrokes = true;
	emit recordingKeystrokes(true);
}

#undef KeyPress

void SSEXEditor::replayKeystrokes()
{
	if(m_bRecordingKeystrokes)
	{
		m_bRecordingKeystrokes = false;
		emit recordingKeystrokes(false);
		return;
	}

	for(SSEXEditorKeystroke *k=m_pKeystrokes->first();k;k=m_pKeystrokes->next())
	{
		QKeyEvent ev(QEvent::KeyPress,k->key,k->ascii,k->state);
		keyPressEvent(&ev);
	}
}

void SSEXEditor::keyPressEvent(QKeyEvent *e)
{
	if(	(e->key() == Qt::Key_Alt) ||
		(e->key() == Qt::Key_Meta) ||
		(e->key() == Qt::Key_Control) ||
		(e->key() == Qt::Key_Shift))
	{
		e->ignore();
		return;
	}

	if(m_bRecordingKeystrokes)
	{
		if(!(((e->key() == Qt::Key_T) || (e->key() == Qt::Key_R)) && (e->state() & ControlButton)))
		{
			SSEXEditorKeystroke * k = new SSEXEditorKeystroke;
			k->ascii = e->ascii();
			k->key = e->key();
			k->state = e->state();
			m_pKeystrokes->append(k);
		}
	}

	m_bCursorOn = true;
	bool bShiftOn = (e->state() & ShiftButton);

	if(e->state() & ControlButton)
	{
		switch(e->key())
		{
			case Qt::Key_Q:     replaceAndFindNext();          e->accept(); return; break;
			case Qt::Key_W:     replaceAndFindNextRegExp();    e->accept(); return; break;
			case Qt::Key_E:     replaceAllInSelection();       e->accept(); return; break;
			case Qt::Key_R:     recordKeystrokes();            e->accept(); return; break;
			case Qt::Key_T:     replayKeystrokes();            e->accept(); return; break;
			case Qt::Key_Y:     goToLine();                    e->accept(); return; break;
			case Qt::Key_U:     unindent();                    e->accept(); return; break;
			case Qt::Key_I:     indent();                      e->accept(); return; break;
			case Qt::Key_O:     commentOut(bShiftOn);          e->accept(); return; break;
			case Qt::Key_P:     removeComment();               e->accept(); return; break;

			case Qt::Key_A:     saveFileAs();                  e->accept(); return; break;
			case Qt::Key_S:     saveFile();                    e->accept(); return; break;
			case Qt::Key_D:     replaceAllInSelectionRegExp(); e->accept(); return; break;
			case Qt::Key_F:     toggleFindWidget();            e->accept(); return; break;
			case Qt::Key_G:     findPrevRegExp();              e->accept(); return; break;
			case Qt::Key_H:     findPrev();                    e->accept(); return; break;
			case Qt::Key_J:     replace();                     e->accept(); return; break;
			case Qt::Key_K:     replaceAll();                  e->accept(); return; break;
			case Qt::Key_L:     replaceAllRegExp();            e->accept(); return; break;

			case Qt::Key_Z:     undo();                        e->accept(); return; break;
			case Qt::Key_X:     cut();                         e->accept(); return; break;
			case Qt::Key_C:     copy();                        e->accept(); return; break;
			case Qt::Key_V:     paste();                       e->accept(); return; break;
			case Qt::Key_B:     findNextRegExp();              e->accept(); return; break;
			case Qt::Key_N:     findNext();                    e->accept(); return; break;
			case Qt::Key_M:     switchMode();                  e->accept(); return; break;
			case Qt::Key_Comma: markBlock();                   e->accept(); return; break;
			case Qt::Key_Period:quickCompletion();             e->accept(); return; break;
		}
		e->ignore();
		return;
	}


	switch(e->key())
	{
		case Qt::Key_Up:        cursorUp(bShiftOn);         e->accept(); return; break;
		case Qt::Key_Down:      cursorDown(bShiftOn);       e->accept(); return; break;
		case Qt::Key_Right:     cursorRight(bShiftOn);      e->accept(); return; break;
		case Qt::Key_Left:      cursorLeft(bShiftOn);       e->accept(); return; break;
		case Qt::Key_PageUp:    cursorPageUp(bShiftOn);     e->accept(); return; break;
		case Qt::Key_PageDown:  cursorPageDown(bShiftOn);   e->accept(); return; break;
		case Qt::Key_Home:      cursorHome(bShiftOn);       e->accept(); return; break;
		case Qt::Key_End:       cursorEnd(bShiftOn);        e->accept(); return; break;

		case Qt::Key_Backspace: backspace();                e->accept(); return; break;
		case Qt::Key_Delete:    del();                      e->accept(); return; break;
		case Qt::Key_Insert:    m_bOverwrite = !m_bOverwrite; e->accept(); return; break;
		case Qt::Key_Return:
		case Qt::Key_Enter:
			if(m_pFindWidget->isVisible())findNext();
			else newLine();
			e->accept();
			return;
		break;
		default:
			if(    ((((unsigned char)e->ascii()) >= 32) && (((unsigned char)e->ascii()) <= 255)) || (e->ascii() == '\t')){
				insertChar(e->ascii());
				e->accept();
				return;
			}
			break;
	}

	e->ignore();
}

#if QT_VERSION >= 300
void SSEXEditor::contextMenuEvent(QContextMenuEvent *e)
{
//	debug("CONTEXT EVENT");
//	contextPopup(e->pos());
	e->ignore();
}
#endif

void SSEXEditor::contextPopup(const QPoint &pos)
{
	int theRow = findRow(pos.y());
	if(theRow < 0)theRow = m_pLines->count() - 1;
	SSEXEditorTextLine * l = m_pLines->at(theRow);
	int charIndex = findCharacterAt((pos.x() - frameWidth()) + xOffset(),l);
	if(charIndex >= l->length)charIndex = l->length - 1;
	contextPopup(l,charIndex);
}

void SSEXEditor::mousePressEvent(QMouseEvent *e)
{
//	debug("MOUSE PRESS");
	if(e->button() & LeftButton){
		m_iCursorRow = findRow(e->pos().y());
		if(m_iCursorRow < 0)m_iCursorRow = m_pLines->count() - 1;
		SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
		m_iCursorPosition = findCharacterAt((e->pos().x() - frameWidth()) + xOffset(),l);
		recalcCursorPosition(l);
		ensureCursorVisible();
		if(m_bHasSelection)clearSelection(false);
		m_selection1 = QPoint(m_iCursorPosition,m_iCursorRow);
		m_selection2 = m_selection1;
		m_mouseAnchor = m_selection1;
		m_lastMousePressCoords = QPoint(e->pos().x() + xOffset(),e->pos().y() + yOffset());
		update();
	} else if(e->button() & MidButton)paste();
	else if(e->button() & RightButton)
	{
//		debug("RIGHT BUTTON");
		contextPopup(e->pos());
	}
//	SSEXTableView::mousePressEvent(e);
}

void SSEXEditor::fillEditPopup(QPopupMenu * p)
{
	QCString tmp;
	SSEXUndo * u = m_pUndoList->last();
	if(u)
	{
		tmp = "&Undo \"";
		tmp.append(u->m_szName);
		tmp.append("\" (CTRL+Z)");
	} else {
		tmp = "Can't undo";
	}


	int id = p->insertItem(QString(tmp.data()),this,SLOT(undo()));
	if(!u)p->setItemEnabled(id,false);
	p->insertSeparator();
	id = p->insertItem("&Copy (CTRL+C)",this,SLOT(copy()));
	if(!m_bHasSelection)p->setItemEnabled(id,false);
	id = p->insertItem("&Paste (CTRL+V)",this,SLOT(paste()));
	QString t = QApplication::clipboard()->text();
	if(t.isNull() || t.isEmpty())p->setItemEnabled(id,false);
	id = p->insertItem("Cu&t (CTRL+X)",this,SLOT(cut()));
	if(!m_bHasSelection)p->setItemEnabled(id,false);

	p->insertSeparator();

	p->insertItem("&Select All",this,SLOT(selectAll()));

	p->insertSeparator();
	p->insertItem("&Go to line (CTRL+Y)",this,SLOT(goToLine()));
	p->insertItem("Go to &beginning",this,SLOT(goToBeginning()));
	p->insertItem("Go to &end",this,SLOT(goToEnd()));
	p->insertSeparator();
	if(m_bHasSelection)
	{
		id = p->insertItem("&Indent selected (CTRL+I)",this,SLOT(indentSelected()));
		id = p->insertItem("&Unindent selected (CTRL+U)",this,SLOT(unindentSelected()));
	} else {
		id = p->insertItem("&Indent current line (CTRL+I)",this,SLOT(indent()));
		id = p->insertItem("&Unindent current line (CTRL+U)",this,SLOT(unindent()));
	}
	p->insertSeparator();
	p->insertItem("&Mark block (CTRL+,)",this,SLOT(markBlock()));
	p->insertItem("&Switch highlight mode (CTRL+M)",this,SLOT(switchMode()));
	p->insertItem("&Complete word (last 10 lines) (CTRL+.)",this,SLOT(quickCompletion()));
	p->insertItem("&Remove trailing CR",this,SLOT(removeCR()));
	p->insertSeparator();

	p->insertItem(m_bRecordingKeystrokes ? "Stop &recording keystrokes (CTRL+R)" : "&Record keystrokes (CTRL+R)",this,SLOT(recordKeystrokes()));
	id = p->insertItem("Replay keys&trokes (CTRL+T)",this,SLOT(replayKeystrokes()));
	if(m_bRecordingKeystrokes || (m_pKeystrokes->count() == 0))p->setItemEnabled(id,false);
}

void SSEXEditor::fillFindPopup(QPopupMenu * p)
{
	p->insertItem(m_pFindWidget->isVisible() ? "Hide &find widget (CTRL+F)" : "&Show &find widget (CTRL+F)",this,SLOT(toggleFindWidget()));
	p->insertSeparator();

	QString toFind = m_pFindWidget->m_pFindStringEdit->text();
	bool bHasToFind = !(toFind.isEmpty() || toFind.isNull());
	int id = p->insertItem("Find &next (CTRL+N)",this,SLOT(findNext()));
	if(!bHasToFind)p->setItemEnabled(id,false);
	id = p->insertItem("Find &prev (CTRL+H)",this,SLOT(findPrev()));
	if(!bHasToFind)p->setItemEnabled(id,false);
	id = p->insertItem("Find next &reg exp (CTRL+B)",this,SLOT(findNextRegExp()));
	if(!bHasToFind)p->setItemEnabled(id,false);
	id = p->insertItem("Find prev reg &exp (CTRL+G)",this,SLOT(findPrevRegExp()));
	if(!bHasToFind)p->setItemEnabled(id,false);
	id = p->insertItem("Rep&lace (CTRL+J)",this,SLOT(replace()));
	if(!m_bHasSelection)p->setItemEnabled(id,false);
	id = p->insertItem("Replace &All (CTRL+K)",this,SLOT(replaceAll()));
	if(!bHasToFind)p->setItemEnabled(id,false);
	id = p->insertItem("Replace All (Re&gexp) (CTRL+L)",this,SLOT(replaceAllRegExp()));
	if(!bHasToFind)p->setItemEnabled(id,false);
	id = p->insertItem("Replace all in &selection (CTRL+E)",this,SLOT(replaceAllInSelection()));
	if(!(bHasToFind && m_bHasSelection))p->setItemEnabled(id,false);
	id = p->insertItem("Replace all &in selection (regexp) (CTRL+D)",this,SLOT(replaceAllInSelectionRegExp()));
	if(!(bHasToFind && m_bHasSelection))p->setItemEnabled(id,false);
}

void SSEXEditor::editPopupAboutToShow()
{
	m_pEditPopup->clear();

	fillEditPopup(m_pEditPopup);
}

void SSEXEditor::findPopupAboutToShow()
{
	m_pFindPopup->clear();

	fillFindPopup(m_pFindPopup);
}

void SSEXEditor::contextPopup(SSEXEditorTextLine *l,int charIndex)
{
	if(m_pContextPopup->isVisible())return;
	m_pContextPopup->clear();

	m_pContextPopup->insertItem("&Edit",m_pEditPopup);
	m_pContextPopup->insertItem("&Find",m_pFindPopup);

	m_pContextPopup->insertSeparator();
	int id = m_pContextPopup->insertItem("&Save (CTRL+S)",this,SLOT(saveFile()));
	if(!m_bModified)m_pContextPopup->setItemEnabled(id,false);
	m_pContextPopup->insertItem("Save &as (CTRL+A)",this,SLOT(saveFileAs()));

	m_pContextPopup->insertSeparator();

	m_pContextPopup->insertItem("Choose font (temporary)",this,SLOT(chooseFont()));

	m_pContextPopup->insertSeparator();

	m_pContextPopup->insertItem("Configure...",this,SLOT(editOptions()));

//#warning "The openinclude stuff should be inside the editor!"
	emit rightClickOnTextRow(this,l->text,charIndex,m_pContextPopup);

	m_pContextPopup->popup(QCursor::pos());
}



void SSEXEditor::dragTimer()
{
	// Send fake mouse events....
	QMouseEvent e(QEvent::MouseMove,mapFromGlobal(QCursor::pos()),LeftButton,LeftButton);
	mouseMoveEvent(&e);
}

void SSEXEditor::mouseMoveEvent(QMouseEvent *e)
{
	if(e->state() & LeftButton)
	{
		QPoint mousePos = e->pos();

		int lastCursorYPosition = m_iCursorRow;

		int rowUnderMouse = findRow(e->pos().y());
		if(mousePos.y() < minViewY())
		{
			if(!m_pDragTimer->isActive())m_pDragTimer->start(SSEX_EDITOR_DRAG_TIMEOUT);
			if(m_iCursorRow > 0)rowUnderMouse = m_iCursorRow - 1;
			else return;
			mousePos.setY(frameWidth() + 1);
		} else if(mousePos.y() > maxViewY())
		{
			if(!m_pDragTimer->isActive())m_pDragTimer->start(SSEX_EDITOR_DRAG_TIMEOUT);
			if(m_iCursorRow < (m_pLines->count() - 1))rowUnderMouse = m_iCursorRow + 1;
			else return;
			mousePos.setY(frameWidth() + viewHeight() - 1);
		} else if(m_pDragTimer->isActive())m_pDragTimer->stop();

		m_iCursorRow = rowUnderMouse;

		SSEXEditorTextLine * l = m_pLines->at(m_iCursorRow);
		m_iCursorPosition = findCharacterAt((mousePos.x() - frameWidth()) + xOffset(),l);
		recalcCursorPosition(l);
		ensureCursorVisible();

		if(!m_bHasSelection)setHasSelection(true);

		lastCursorYPosition = ((lastCursorYPosition * cellHeight()) - yOffset()) + frameWidth();
		
		if(m_mouseAnchor.y() == m_iCursorRow)
		{
			if(m_lastMousePressCoords.x() <= mousePos.x())
			{
				m_selection1 = m_mouseAnchor;
				m_selection2 = QPoint(m_iCursorPosition,m_iCursorRow);
			} else {
				m_selection2 = m_mouseAnchor;
				m_selection1 = QPoint(m_iCursorPosition,m_iCursorRow);
			}
		} else {
			if(m_lastMousePressCoords.y() <= (mousePos.y() + yOffset()))
			{
				m_selection1 = m_mouseAnchor;
				m_selection2 = QPoint(m_iCursorPosition,m_iCursorRow);
			} else {
				m_selection2 = m_mouseAnchor;
				m_selection1 = QPoint(m_iCursorPosition,m_iCursorRow);
			}
		}

		int cH = cellHeight() + 1;
		if(lastCursorYPosition > mousePos.y())update(QRect(0,mousePos.y() - cH,width(),(lastCursorYPosition - mousePos.y()) + (cH << 1)));
		else update(QRect(0,lastCursorYPosition - cH,width(),(mousePos.y() - lastCursorYPosition) + (cH << 1)));
	}
}

void SSEXEditor::mouseReleaseEvent(QMouseEvent *)
{
	if(m_pDragTimer->isActive())m_pDragTimer->stop();
}

void SSEXEditor::updateCellSize()
{
	setCellHeight(m_iFontLineSpacing);
	setCellWidth(m_iMaxTextWidth + (SSEX_EDITOR_BORDER * 2));
	m_pMemBuffer->resize(cellWidth() + 1,cellHeight() + 1);
}


void SSEXEditor::updateFontDependantVariables()
{
	QFontMetrics fm(font());
	int spaceSize = fm.width(' ');
	m_iTabsNumPixels = spaceSize * g_pSSEXOptions->uTabsNumSpaces;
	for(int i=0;i<256;i++)m_iCharWidth[i] = fm.width((char)i);
	m_iFontAscent = fm.ascent();
	m_iFontLineSpacing = fm.lineSpacing();
	updateCellSize();
}

void SSEXEditor::chooseFont()
{
	bool bOk;
	QFont f = QFontDialog::getFont(&bOk,font(),this);
	if(bOk)setFont(f);
}

void SSEXEditor::setFont(const QFont &f)
{
	SSEXTableView::setFont(f);
	updateFontDependantVariables();
	if(m_pLines->count() > m_iCursorRow)
	{
		SSEXEditorTextLine *l = m_pLines->at(m_iCursorRow);
		if(l)recalcCursorPosition(l);
	}
	update();
}

void SSEXEditor::updateOptions()
{
	initializeCurrentMode();
	update();
}

void SSEXEditor::clear()
{
	setText("");
	setModified(true);
}

int SSEXEditor::getTextWidthWithTabsForCursorPosition(const char *text,int cursorPos)
{
	int totWidth = 0;
	int lastTabStop = 0;
	while(*text && cursorPos)
	{
		if(*text == '\t')
		{
			while(lastTabStop <= totWidth)lastTabStop += m_iTabsNumPixels;
			totWidth = lastTabStop;
		} else totWidth += m_iCharWidth[(unsigned char)(*text)];
		text++;
		if(!(--cursorPos))return totWidth;
	}
	return totWidth;
}

int SSEXEditor::findCharacterAt(int xPositionInCell,SSEXEditorTextLine * l)
{
	const char *aux = l->text.data();
	int curXPos = SSEX_EDITOR_BORDER;
	int lastTabStop = SSEX_EDITOR_BORDER;
	while(*aux)
	{
		if(*aux == '\t')
		{
			while(lastTabStop <= curXPos)lastTabStop += m_iTabsNumPixels;
			if(xPositionInCell < (curXPos + ((lastTabStop - curXPos) >> 1)))return (aux - l->text.data());
			else {
				curXPos = lastTabStop;
				aux++;
			}
		} else {
			if(xPositionInCell < (curXPos + (m_iCharWidth[(unsigned char)*aux] >> 1)))return (aux - l->text.data());
			else {
				curXPos += m_iCharWidth[(unsigned char)*aux];
				aux++;
			}
		}
	}
	return l->length;
}

int SSEXEditor::getTextWidthWithTabs(const char *text)
{
	int totWidth = 0;
	int lastTabStop = 0;
	while(*text)
	{
		if(*text == '\t')
		{
			while(lastTabStop <= totWidth)lastTabStop += m_iTabsNumPixels;
			totWidth = lastTabStop;
		} else totWidth += m_iCharWidth[(unsigned char)*text];
		text++;
	}
	return totWidth;
}

void SSEXEditor::editOptions()
{
	g_pSSEXOptions->edit();
}

void SSEXEditor::setText(const QCString &text)
{
	m_iCursorRow = 0;
	m_iCursorPosition = 0;

	m_bHasSelection = false;
	m_bHasBlockMark = false;

	m_pLines->clear();

	const char *aux = text.data();

	if(!aux)aux = "";

	do {
		const char * begin = aux;
		while(*aux && (*aux != '\n'))aux++;
		SSEXEditorTextLine * l = new SSEXEditorTextLine;
		l->length = aux - begin;
		l->text = QCString(begin,l->length + 1);
		if(*aux)aux++;
		l->width = getTextWidthWithTabs(l->text.data());
		l->flags = 0;
		m_pLines->append(l);
	} while(*aux);

	setNumRows(m_pLines->count());

	initializeCurrentMode();

	updateMaxTextWidth();
	updateCellSize();
	recalcCursorPosition(m_pLines->first());

	setOffset(0,0);

	clearUndoList();

	update();
}

bool SSEXEditor::loadFile(const char *filename)
{
	QFile f(filename);
	if(!f.open(IO_ReadOnly))return false;
	QCString buffer;
	buffer.resize(f.size() + 1);
	int readed = f.readBlock(buffer.data(),f.size());
	*(buffer.data() + readed) = '\0';
	f.close();
	QCString fExtension = filename;
	int idx = fExtension.findRev('.');

	if(readed != buffer.length())
	{
		// Binary file!
	}

	// Lookup #!/bin/something magic files!

	ColorMode mode = Normal;
	if(idx != -1)
	{
		fExtension.remove(0,idx+1);
		if(strcasecmp(fExtension.data(),"kvs") == 0)mode = Perl;
	}
	setMode(mode);
	setText(buffer);
	m_szFileName = filename;
	setModified(false);
	emit fileNameChanged(this,m_szFileName);
	return true;
}

bool SSEXEditor::saveFile()
{
	if(m_szFileName.isEmpty() || m_szFileName.isNull())return saveFileAs();
	else return saveFile(m_szFileName.data());
}

bool SSEXEditor::saveFile(const char *filename)
{
	QFile f(filename);
	if(!f.open(IO_WriteOnly)){
		QMessageBox::warning(this,"Warning - KVIrc","Cannot open the file for writing.\nSave failed.",QMessageBox::Ok | QMessageBox::Default,0);
		return false;
	}
	int i = 0;
	int lastProg = -1;
	int progress;
	if(m_pLines->count() == 0){
		emit saveProgress(m_szFileName,100);
	} else {
		emit saveProgress(m_szFileName,0);
		for(SSEXEditorTextLine *l=m_pLines->first();l;l=m_pLines->next())
		{
			__assert(l->text.length() == l->length);
//#ifdef __DEBUG
//			if(l->text.length() != l->length)debug("WARNING : Wrong line size (%d) , should be %d for (%s)",l->length,l->text.length(),l->text.data());
//#endif
			if((f.writeBlock(l->text.data(),l->length) != l->length) || (f.writeBlock("\n",1) != 1))
			i++;
			progress = (i * 100) / m_pLines->count();
			if(progress != lastProg){
				emit saveProgress(m_szFileName,progress);
				lastProg = progress;
			}
		}
	}
	f.close();
	if(strcmp(m_szFileName.data(),filename) != 0){
		m_szFileName = filename;
		emit fileNameChanged(this,m_szFileName);
	}
	setModified(false);
	emit saved(m_szFileName);
	return true;
}

bool SSEXEditor::saveFileAs()
{
	QString newName = QFileDialog::getSaveFileName(m_szFileName,QString::null,0);
	if(newName.isNull())return false; //Aborted
	QFileInfo fi(newName);
	if(fi.exists())
	{
		QString msg;
		msg.sprintf("The file %s already exists.\nDo you wish to overwrite it?",newName.ascii());
		int result = QMessageBox::information(this,"Warning - KVIrc",msg,QMessageBox::Yes,QMessageBox::No | QMessageBox::Default);
		if(result != QMessageBox::Yes)return false; // Aborted
	}
	return saveFile(newName.ascii());
}

void SSEXEditor::initializeCurrentMode()
{
	QPalette pal = palette();
	switch(m_mode)
	{
		case Perl:
			setBackgroundColor(g_pSSEXOptions->clrPerlExtBackground);
			setBackgroundMode(NoBackground);
			setFont(g_pSSEXOptions->fntPerl);
		break;
		default:
			setBackgroundColor(g_pSSEXOptions->clrTxtExtBackground);
			setBackgroundMode(NoBackground);
			setFont(g_pSSEXOptions->fntTxt);
		break;
	}
}


#ifdef KVIRC_SSEX
	#include "m_editor.moc"
#else
	#include "moc_editor.cpp"
#endif
